# Copyright (c) 2020 kitsune.ONE team.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import bpy
import json
import math
import mathutils  # make sure to "import bpy" before
import struct

from bpy_extras.io_utils import ExportHelper
from typing import Set, cast

from metadrive.third_party.kitsunetsuki.base.armature import get_armature
from metadrive.third_party.kitsunetsuki.base.context import Mode
from metadrive.third_party.kitsunetsuki.base.collections import get_object_collection
from metadrive.third_party.kitsunetsuki.base.matrices import (
    get_bone_matrix, get_object_matrix, get_inverse_bind_matrix, matrix_to_list, quat_to_list
)
from metadrive.third_party.kitsunetsuki.base.objects import (
    is_collision, is_object_visible, get_object_properties, set_active_object
)

from metadrive.third_party.kitsunetsuki.exporter.base import Exporter

from . import spec
from .buffer import GLTFBuffer
from .animation import AnimationMixin
from .geom import GeomMixin
from .material import MaterialMixin
from .vertex import VertexMixin
from .texture import TextureMixin


class GLTFExporter(AnimationMixin, GeomMixin, MaterialMixin, VertexMixin, TextureMixin, Exporter):
    """
    BLEND to GLTF converter.
    """
    def __init__(self, args):
        super().__init__(args)
        self._output = args.output or args.inputs[0].replace('.blend', '.gltf')
        self._z_up = getattr(args, 'z_up', False)
        self._pose_freeze = getattr(args, 'pose_freeze', False)
        self._split_primitives = getattr(args, 'split_primitives', False)
        self._norm_weights = getattr(args, 'normalize_weights', False)

        if self._z_up:
            self._matrix = mathutils.Matrix((
                (1.0, 0.0, 0.0),
                (0.0, 1.0, 0.0),
                (0.0, 0.0, 1.0),
            ))
            self._matrix_inv = self._matrix
        else:
            # Mat3.convert_mat(CS_yup_right, CS_zup_right)
            self._matrix = mathutils.Matrix((
                (1.0, 0.0, 0.0),
                (0.0, 0.0, 1.0),
                (0.0, -1.0, 0.0),
            ))

            # Mat3.convert_mat(CS_zup_right, CS_yup_right)
            self._matrix_inv = mathutils.Matrix((
                (1.0, 0.0, 0.0),
                (0.0, 0.0, -1.0),
                (0.0, 1.0, 0.0),
            ))

    def _transform(self, x):
        """
        Transform matrix/vector using axis conversion matrices.
        """
        if self._z_up:
            return x
        else:
            return (self._matrix.to_4x4() @ x @ self._matrix_inv.to_4x4())

    def _freeze(self, matrix):
        """
        Freezes matrix. Removes rotation and scale.
        """
        pos = matrix.to_translation()
        return mathutils.Matrix.Translation(pos).to_4x4()

    def make_root_node(self):
        gltf_node = {
            'asset': {
                'generator': (
                    'KITSUNETSUKI Asset Tools by kitsune.ONE - '
                    'https://github.com/kitsune-ONE-team/KITSUNETSUKI-Asset-Tools'
                ),
                'version': '2.0',
            },
            'extensions': {
                # 'BP_physics_engine': {'engine': 'bullet'},
            },
            'extensionsUsed': [],
            'scene': 0,
            'scenes': [{
                'name': 'Scene',
                'nodes': [],
            }],
            'nodes': [],
            'meshes': [],
            'materials': [{
                # skips panda warnings
                'name': 'GLTF_DEFAULT_MATERIAL',
            }],
            'animations': [],
            'skins': [],
            'textures': [],  # links to samplers-images pair
            'samplers': [],
            'images': [],
            'accessors': [],
            'bufferViews': [],
            'buffers': [],
        }

        if self._z_up:
            gltf_node['extensionsUsed'].append('BP_zup')

        return gltf_node

    def _add_child(self, parent_node, child_node):
        self._root['nodes'].append(child_node)
        node_id = len(self._root['nodes']) - 1

        if 'scenes' in parent_node:
            self._root['scenes'][0]['nodes'].append(node_id)
        else:
            if 'children' not in parent_node:
                parent_node['children'] = []
            parent_node['children'].append(node_id)

    def _setup_node(self, node, obj=None, can_merge=False):
        if obj is None:
            return

        armature = get_armature(obj)
        obj_matrix = self._transform(get_object_matrix(obj, armature=armature))

        # get custom object properties
        obj_props = get_object_properties(obj)

        if not can_merge and not armature:
            if self._geom_scale == 1:
                node.update(
                    {
                        'rotation': quat_to_list(obj_matrix.to_quaternion()),
                        'scale': list(obj_matrix.to_scale()),
                        'translation': list(obj_matrix.to_translation()),
                        # 'matrix': matrix_to_list(obj_matrix),
                    }
                )
            else:
                x, y, z = list(obj_matrix.to_translation())
                x *= self._geom_scale
                y *= self._geom_scale
                z *= self._geom_scale
                node.update(
                    {
                        'rotation': quat_to_list(obj_matrix.to_quaternion()),
                        'scale': list(obj_matrix.to_scale()),
                        'translation': [x, y, z],
                    }
                )

        # setup collisions
        if not can_merge and is_collision(obj) and obj_props.get('type') != 'Portal':
            collision = {}
            node['extensions'] = {
                'BLENDER_physics': collision,
            }

            # collision shape
            shape = {
                'shapeType': obj.rigid_body.collision_shape,
                'boundingBox': [obj.dimensions[i] / obj_matrix.to_scale()[i] * self._geom_scale for i in range(3)],
            }
            if obj.rigid_body.collision_shape == 'MESH':
                shape['mesh'] = node.pop('mesh')

            collision['collisionShapes'] = [shape]
            collision['static'] = obj.rigid_body.type == 'PASSIVE'

            # don't actually collide (ghost)
            if (not obj.collision or not obj.collision.use):
                collision['intangible'] = True

            if self._set_origin:
                obj.select_set(state=True)
                set_active_object(obj)
                x1, y1, z1 = obj.location
                # set origin to the center of bounds
                bpy.ops.object.origin_set(type='ORIGIN_GEOMETRY', center='BOUNDS')
                x2, y2, z2 = obj.location
                if 'extras' not in node:
                    node['extras'] = {}
                node['extras']['origin'] = [x2 - x1, y2 - y1, z2 - z1]

        # setup custom properties with tags
        if (obj_props or can_merge) and 'extras' not in node:
            node['extras'] = {}

        for k, v in obj_props.items():
            if node['extras'].get(k):  # tag exists
                tag = node['extras'].get(k)

            if type(v) in (tuple, list, dict):
                tag = json.dumps(v)
            else:
                tag = '{}'.format(v)

            node['extras'][k] = tag

        # if can_merge and 'type' not in obj_props:
        if can_merge:
            node['extras']['type'] = 'Merged'

    def make_empty(self, parent_node, obj):
        gltf_node = {
            'name': obj.name,
        }

        self._setup_node(gltf_node, obj)
        self._add_child(parent_node, gltf_node)

        return gltf_node

    def make_armature(self, parent_node, armature):
        channel = self._buffer.add_channel(
            {
                'componentType': spec.TYPE_FLOAT,
                'type': 'MAT4',
                'extras': {
                    'reference': 'inverseBindMatrices',
                },
            }
        )

        gltf_armature = {
            'name': armature.name,
        }
        gltf_skin = {
            'name': armature.name,
            'joints': [],
            'inverseBindMatrices': channel['bufferView'],
        }
        self._root['skins'].append(gltf_skin)

        set_active_object(armature)

        bone_tails_local = {}
        for bone_name, bone in armature.data.bones.items():
            bone_tails_local[bone_name] = bone.tail

        bone_tails_off = {}
        with Mode('EDIT'):
            for bone_name, bone in armature.data.edit_bones.items():
                bone_tails_off[bone_name] = bone.tail - bone.head

        if self._pose_freeze:
            # set_active_object(armature)

            # disconnect all bones
            with Mode('EDIT'):
                for bone_name, bone in armature.data.edit_bones.items():
                    bone.use_connect = False

            # reset bones rotation
            with Mode('EDIT'):
                for bone_name, bone in armature.data.edit_bones.items():
                    bone.roll = 0
                    bone.length = 10
                    bone.tail = bone.head + mathutils.Vector((0, bone.length, 0))
                    bone.roll = 0

        # create joint nodes
        gltf_joints = {}
        for bone_name, bone in armature.data.bones.items():
            bone_matrix = self._transform(get_bone_matrix(bone, armature))
            bone_tail_matrix = self._transform(mathutils.Matrix.Translation(bone_tails_local[bone_name]))

            if self._pose_freeze:
                bone_matrix = self._freeze(bone_matrix)
                bone_tail_matrix = self._transform(mathutils.Matrix.Translation(bone_tails_off[bone_name]))

            # print(mathutils.Matrix.Translation(bone_tails[bone_name]).to_translation())
            # print((mathutils.Matrix.Translation(bone_tails[bone_name]) @
            #       bone.matrix.to_4x4() @
            #       bone.matrix.to_4x4().inverted()).to_translation())

            gltf_joint = {
                'name': bone_name,
                'rotation': quat_to_list(bone_matrix.to_quaternion()),
                'scale': list(bone_matrix.to_scale()),
                'translation': list(bone_matrix.to_translation()),
                'extras': {
                    'tail': {
                        'translation': list(bone_tail_matrix.to_translation()),
                    }
                }
            }
            gltf_joints[bone_name] = gltf_joint

        # add joints to skin
        for bone_name in sorted(gltf_joints.keys()):
            bone = armature.data.bones[bone_name]
            if bone.parent:  # attach joint to parent joint
                self._add_child(gltf_joints[bone.parent.name], gltf_joints[bone.name])
            else:  # attach joint to armature
                # gltf_skin['skeleton'] = len(self._root['nodes']) - 1
                self._add_child(gltf_armature, gltf_joints[bone.name])

            ib_matrix = self._transform(get_inverse_bind_matrix(bone, armature))
            if self._pose_freeze:
                ib_matrix = self._freeze(ib_matrix)

            self._buffer.write(gltf_skin['inverseBindMatrices'], *matrix_to_list(ib_matrix))
            gltf_skin['joints'].append(len(self._root['nodes']) - 1)

        self._setup_node(gltf_armature, armature)
        self._add_child(parent_node, gltf_armature)

        # no meshes or animation only
        if (not list(filter(is_object_visible, armature.children)) or self._export_type == 'animation'):
            gltf_child_node = {
                'name': '{}_EMPTY'.format(armature.name),
                'skin': len(self._root['skins']) - 1,
            }
            self._add_child(gltf_armature, gltf_child_node)

        return gltf_armature

    def _make_node_mesh(self, parent_node, name, obj=None, can_merge=False):
        """
        Make glTF-node - glTF-mesh pair for chosen Blender object.
        """
        gltf_node = {
            'name': name,
            'extras': {},
        }

        gltf_mesh = None
        need_mesh = False
        if can_merge:
            need_mesh = True
        else:
            if obj.rigid_body is None:
                need_mesh = True
            elif obj.rigid_body.collision_shape == 'MESH':
                need_mesh = True

        if need_mesh:
            gltf_mesh = {
                'name': name,
                'primitives': [],
                'extras': {
                    'targetNames': [],
                },
            }
            self._root['meshes'].append(gltf_mesh)
            gltf_node['mesh'] = len(self._root['meshes']) - 1

        armature = obj and get_armature(obj)
        if armature:
            for i, gltf_skin in enumerate(self._root['skins']):
                if gltf_skin['name'] == armature.name:
                    gltf_node['skin'] = i
                    break

        self._setup_node(gltf_node, obj, can_merge=can_merge)
        self._add_child(parent_node, gltf_node)

        if obj.type == 'MESH':
            for material in obj.data.materials:
                if material and material.node_tree:  # not and empty slot and have nodes tree
                    for node in material.node_tree.nodes:
                        if node.type == 'ATTRIBUTE':
                            gltf_node['extras'][node.attribute_type.lower()] = node.attribute_name

        return gltf_node, gltf_mesh

    def make_mesh(self, parent_node, obj):
        """
        Make mesh-type object.
        """
        gltf_node = None

        # merged nodes
        # if self.can_merge(obj):
        if False:
            collection = get_object_collection(obj)

            for child in self._root['nodes']:
                if child['name'] == collection.name:
                    # got existing glTF node
                    gltf_node = child

                    mesh_id = gltf_node['mesh']
                    # got existing glTF mesh
                    gltf_mesh = self._root['meshes'][mesh_id]
                    break
            else:  # glTF-node - glTF-mesh pair not found
                # create new pair
                gltf_node, gltf_mesh = self._make_node_mesh(parent_node, collection.name, obj, can_merge=True)

            if gltf_mesh:
                self.make_geom(gltf_node, gltf_mesh, obj, can_merge=True)

        # separate nodes
        # if not self.can_merge(obj) or self._keep:
        if True:
            obj_props = get_object_properties(obj)
            if obj_props.get('type') == 'Portal':
                vertices = [list(vertex.co) for vertex in obj.data.vertices]
                gltf_node = {
                    'name': obj.name,
                    'extras': {
                        'vertices': json.dumps(vertices)
                    },
                }
                self._setup_node(gltf_node, obj, can_merge=False)
                self._add_child(parent_node, gltf_node)
                return gltf_node

            gltf_node, gltf_mesh = self._make_node_mesh(parent_node, obj.name, obj, can_merge=False)

            if gltf_mesh:
                self.make_geom(gltf_node, gltf_mesh, obj, can_merge=False)

        return gltf_node

    def make_light(self, parent_node, obj):
        """
        Make light-type object.
        """
        LIGHT_TYPES = {
            'POINT': 'PointLight',
            'SPOT': 'SpotLight',
        }

        gltf_light = {
            'name': obj.name,
            'extras': {
                'type': 'Light',
                'light': LIGHT_TYPES[obj.data.type],
                'color': json.dumps(tuple(obj.data.color)),
                'scale': json.dumps(tuple(obj.scale)),
                'energy': '{:.3f}'.format(obj.data.energy),
                'far': '{:.3f}'.format(obj.data.shadow_soft_size),
            },
        }

        if obj.data.type == 'SPOT':
            gltf_light['extras']['fov'] = '{:.3f}'.format(math.degrees(obj.data.spot_size))

        self._setup_node(gltf_light, obj)
        self._add_child(parent_node, gltf_light)

        return gltf_light

    def convert(self):
        self._buffer = GLTFBuffer(self._output)
        root = super().convert()
        return root, self._buffer

    def write(self, root, output, is_binary=False):
        if is_binary:
            with open(output, 'wb') as f:  # binary mode
                chunk1 = self._buffer.export(root)  # export buffer first because it updates gltf data
                chunk0 = json.dumps(root, indent=4).encode()  # export gltf data
                while len(chunk0) % 4:
                    chunk0 += b' '
                while len(chunk1) % 4:
                    chunk1 += b'\x00'

                # write global headers
                f.write(b'glTF')  # header
                f.write(struct.pack('<I', 2))  # version
                size = (
                    4 + 4 + 4 +  # global headers
                    4 + 4 + len(chunk0) +  # chunk0 + headers
                    4 + 4 + len(chunk1)
                )  # chunk1 + headers
                f.write(struct.pack('<I', size))  # full size

                # write chunk0 with headers
                f.write(struct.pack('<I', len(chunk0)))
                f.write(b'JSON')
                f.write(chunk0)

                # write chunk1 with headers
                f.write(struct.pack('<I', len(chunk1)))
                f.write(b'BIN\0')
                f.write(chunk1)

        else:
            with open(output, 'w') as f:  # text mode
                json.dump(root, f, indent=4)


class GLTFExporterOperator(bpy.types.Operator, ExportHelper):
    bl_idname = 'scene.gltf'
    bl_label = 'Export glTF'
    bl_description = 'Export glTF'
    bl_options = {'REGISTER', 'UNDO'}

    filename_ext = '.gltf'
    filter_glob: bpy.props.StringProperty(default='*.gltf', options={'HIDDEN'})

    def execute(self, context: bpy.types.Context):
        if not self.filepath:
            return {'CANCELLED'}

        class Args(object):
            inputs = []
            output = self.filepath
            export = 'all'
            render = 'default'
            exec = None
            action = None
            speed = None
            scale = None
            merge = None
            keep = None
            no_extra_uv = None
            no_materials = None
            no_textures = None
            empty_textures = None
            set_origin = None
            normalize_weights = None

        args = Args()
        e = GLTFExporter(args)
        out, buf = e.convert()

        if args.output.endswith('.gltf'):
            # write buffer into separate file
            buf.export(out, args.output.replace('.gltf', '.bin'))
            # write glTF data
            e.write(out, args.output, is_binary=False)

        else:
            # write glTF data with embedded buffer
            e.write(out, args.output, is_binary=True)

        # re-open current file
        bpy.ops.wm.open_mainfile(filepath=bpy.data.filepath)

        return {"FINISHED"}

    def invoke(self, context, event):
        return cast(Set[str], ExportHelper.invoke(self, context, event))

    def draw(self, context):
        pass


def export(export_op, context):
    export_op.layout.operator(GLTFExporterOperator.bl_idname, text='glTF using KITSUNETSUKI Asset Tools (.gltf)')



from typing import Union, Tuple

from direct.showbase import ShowBase
from panda3d.bullet import BulletPlaneShape, BulletRigidBodyNode, BulletDebugNode
from panda3d.core import Vec3, NodePath, LineSegs

from metadrive.component.algorithm.BIG import NextStep
from metadrive.component.map.base_map import BaseMap
from metadrive.constants import BKG_COLOR
from metadrive.constants import CollisionGroup
from metadrive.engine.core.physics_world import PhysicsWorld


class TestBlock(ShowBase.ShowBase):
    def __init__(self, debug=False, window_type="onscreen"):
        self.debug = debug
        super(TestBlock, self).__init__(windowType=window_type)
        self.setBackgroundColor(BKG_COLOR)
        if window_type != "none":
            self.setFrameRateMeter(True)
            self.cam.setPos(0, 0, 300)
            self.cam.lookAt(0, 0, 0)
        self.use_render_pipeline = False
        self.map = None
        self.worldNP = None
        self.world = None
        self.physics_world = None
        self.debugNP = None
        self.groundNP = None
        self.setup()
        self.taskMgr.add(self.update, 'updateWorld')
        self.taskMgr.add(self.analyze, "analyze geom node")
        self.add_block_func = None  # function pointer
        self.last_block = None
        self.block_index = 1
        self.big = None
        self.accept("f4", self.render.analyze)
        self.accept("4", self.render.analyze)
        self.accept("f1", self.toggleDebug)
        self.accept("1", self.toggleDebug)
        self.vehicle = None
        self.inputs = None

    def toggleDebug(self):
        if self.debugNP is None:
            debugNode = BulletDebugNode('Debug')
            debugNode.showWireframe(True)
            debugNode.showConstraints(True)
            debugNode.showBoundingBoxes(False)
            debugNode.showNormals(True)
            debugNP = self.render.attachNewNode(debugNode)
            self.physics_world.static_world.setDebugNode(debugNP.node())
            self.debugNP = debugNP
        if self.debugNP.isHidden():
            self.debugNP.show()
        else:
            self.debugNP.hide()

    def vis_big(self, big):
        # self.cam.setPos(200, 700, 1000)
        # self.cam.lookAt(200, 700, 0)
        self.big = big
        self.big.next_step = NextStep.forward
        self.task_mgr.add(self.big_algorithm, "start_big")

    def big_algorithm(self, task):
        if self.big.big_helper_func():
            return task.done
        else:
            return task.cont

    def setup(self):
        self.worldNP = self.render.attachNewNode('World')
        self.world = PhysicsWorld()
        self.physics_world = self.world

        # Ground (static)
        shape = BulletPlaneShape(Vec3(0, 0, 1), 0)
        self.groundNP = self.worldNP.attachNewNode(BulletRigidBodyNode('Ground'))
        self.groundNP.node().addShape(shape)
        self.groundNP.setPos(0, 0, 0)
        self.groundNP.setCollideMask(CollisionGroup.AllOn)
        self.world.dynamic_world.attachRigidBody(self.groundNP.node())

    def update(self, task):
        dt = 1 / 60
        # self.world.doPhysics(dt)
        self.world.dynamic_world.doPhysics(dt, 1, dt)
        return task.cont

    def analyze(self, task):
        self.render.analyze()
        return task.done

    def test_reset(self, big_type, para):
        """
        For reset test and map storing
        """
        self.map = BaseMap(self, {"type": big_type, "config": para})
        self.accept("c", self.clear)
        self.accept("a", self.re_add)

    def clear(self):
        self.map.unload_map(self)

    def re_add(self):
        self.map.load_map(self)

    def draw_line_3d(self, start_p: Union[Vec3, Tuple], end_p: Union[Vec3, Tuple], color, thickness: float):
        line_seg = LineSegs("interface")
        line_seg.setColor(*color)
        line_seg.moveTo(start_p)
        line_seg.drawTo(end_p)
        line_seg.setThickness(thickness)
        np = NodePath(line_seg.create(False))
        np.reparentTo(self.render)
        # TODO(PZH): NodePath is not registered.

    def show_bounding_box(self, road_network):
        bound_box = road_network.get_bounding_box()
        points = [(x, y) for x in bound_box[:2] for y in bound_box[2:]]
        for k, p in enumerate(points[:-1]):
            for p_ in points[k + 1:]:
                self.draw_line_3d((*p, 2), (*p_, 2), (1, 0., 0., 1), 2)


if __name__ == "__main__":
    TestBlock = TestBlock()
    TestBlock.run()



"""

Uses mitsuba to generate the environment brdf

"""

from __future__ import print_function

import os
import sys
import math
from panda3d.core import PNMImage, load_prc_file_data, Vec3

load_prc_file_data("", "notify-level error")
load_prc_file_data("", "notify-level-pnmimage error")

configs = {
    "normal": {
        "out_dir": "slices",
        "out_name": "env_brdf_{}.png",
        "template_suffix": "",
        "sequence": xrange(7),
        "samples": 32,
    },
    "metallic": {
        "out_dir": "slices_metal",
        "out_name": "env_brdf.png",
        "template_suffix": "-metal",
        "sequence": [1],
        "samples": 32,
    },
    "clearcoat": {
        "out_dir": "slices_coat",
        "out_name": "env_brdf.png",
        "template_suffix": "-coat",
        "sequence": [1],
        "samples": 2048,
    }
}

# configs_to_run = ["normal", "metallic", "clearcoat"]
configs_to_run = ["clearcoat"]

for config_name in configs_to_run:

    config = configs[config_name]

    if not os.path.isdir(config["out_dir"]):
        os.makedirs(config["out_dir"])

    for pass_index in config["sequence"]:

        ior = 1.01 + 0.2 * pass_index

        dest_size = 512
        dest_h = 32
        dest = PNMImage(dest_size, dest_h)

        # run mitsuba
        print("Running mitsuba for ior =", ior, "( index =", pass_index, ")")
        with open("res/scene" + config["template_suffix"] + ".templ.xml", "r") as handle:
            content = handle.read()

        content = content.replace("%IOR%", str(ior))
        content = content.replace("%SAMPLES%", str(config["samples"]))

        with open("res/scene.xml", "w") as handle:
            handle.write(content)

        os.system("run_mitsuba.bat")

        img = PNMImage("scene.png")
        source_w = img.get_x_size()

        print("Converting ..")

        indices = []
        nxv_values = []

        # Generate nonlinear NxV sequence
        for i in xrange(source_w):
            v = 1 - i / float(source_w)
            NxV = math.sqrt(1 - v * v)
            nxv_values.append(NxV)

        # Generate lerp indices and weights
        for x in xrange(dest_size):
            NxV = (x) / float(dest_size)
            index = 0
            for i, s_nxv in enumerate(reversed(nxv_values)):
                if NxV >= s_nxv:
                    index = i
                    break
            index = len(nxv_values) - index - 1
            next_index = index + 1 if index < dest_size - 1 else index

            curr_nxv = nxv_values[index]
            next_nxv = nxv_values[next_index]

            lerp_factor = (NxV - curr_nxv) / max(1e-10, abs(next_nxv - curr_nxv))
            lerp_factor = max(0.0, min(1.0, lerp_factor))
            indices.append((index, next_index, lerp_factor))

        # Generate the final linear lut using the lerp weights
        for y in xrange(dest_h):
            for x in xrange(dest_size):
                curr_i, next_i, lerp = indices[x]
                curr_v = img.get_xel(curr_i, y)
                next_v = img.get_xel(next_i, y)
                dest.set_xel(x, y, curr_v * (1 - lerp) + next_v * lerp)

        out_name = config["out_name"].replace("{}", str(pass_index))
        dest.write(config["out_dir"] + "/" + out_name)

try:
    os.remove("scene.png")
except:
    pass



"""

RenderPipeline

Copyright (c) 2014-2016 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""
from metadrive.render_pipeline.rplibs.six import iteritems

from panda3d.core import PTAFloat, PTALVecBase3f, PTALMatrix4f, PTALVecBase2f
from panda3d.core import PTALVecBase4f, PTALMatrix3f, PTAInt, TypeRegistry, PTALVecBase2i

from metadrive.render_pipeline.rpcore.rpobject import RPObject


class SimpleInputBlock(RPObject):
    """ Simplest possible uniform buffer which just stores a set of values.
    This does not use any fancy uniform buffer objects under the hood, and
    instead just sets every value as a shader input. """
    def __init__(self, name):
        """ Creates the ubo with the given name """
        RPObject.__init__(self)
        self._inputs = {}
        self.name = name

    def add_input(self, name, value):
        """ Adds a new input to the UBO """
        self._inputs[self.name + "." + name] = value

    def bind_to(self, target):
        """ Binds the UBO to a target """
        target.set_shader_inputs(**self._inputs)


class GroupedInputBlock(RPObject):
    """ Grouped uniform buffer which either uses PointerToArray's to efficiently
    store and update the shader inputs, or in case of uniform buffer object (UBO)
    support, uses these to pass the inputs to the shaders. """

    # Keeps track of the global allocated input blocks to be able to assign
    # a unique binding to all of them
    UBO_BINDING_INDEX = 0

    # Mapping of the pta-types to glsl types, and vice versa
    PTA_MAPPINGS = {
        PTAInt: "int",
        PTAFloat: "float",
        PTALVecBase2f: "vec2",
        PTALVecBase2i: "ivec2",
        PTALVecBase3f: "vec3",
        PTALVecBase4f: "vec4",
        PTALMatrix3f: "mat3",
        PTALMatrix4f: "mat4",
    }

    def __init__(self, name):
        """ Constructs the input block with a given name """
        RPObject.__init__(self)
        self.ptas = {}
        self._inputs = {}
        self.name = name
        self.use_ubo = bool(TypeRegistry.ptr().find_type("GLUniformBufferContext"))

        # Acquire a unique index for each UBO to store its binding
        self.bind_id = GroupedInputBlock.UBO_BINDING_INDEX
        GroupedInputBlock.UBO_BINDING_INDEX += 1

        if self.bind_id == 0:
            # Only output the bind support debug output once (for the first ubo)
            self.debug("Native UBO support =", self.use_ubo)

    def register_pta(self, name, input_type):
        """ Registers a new input, type should be a glsl type """
        pta = self.glsl_type_to_pta(input_type).empty_array(1)
        self.ptas[name] = pta
        if self.use_ubo:
            self._inputs[self.name + "_UBO." + name] = pta
        else:
            self._inputs[self.name + "." + name] = pta

    def pta_to_glsl_type(self, pta_handle):
        """ Converts a PtaXXX to a glsl type """
        for pta_type, glsl_type in iteritems(GroupedInputBlock.PTA_MAPPINGS):
            if isinstance(pta_handle, pta_type):
                return glsl_type
        self.error("Unrecognized PTA type:", pta_handle)

    def glsl_type_to_pta(self, glsl_type):
        """ Converts a glsl type to a PtaXXX type """
        for key, val in iteritems(GroupedInputBlock.PTA_MAPPINGS):
            if val == glsl_type:
                return key
        self.error("Could not resolve GLSL type:", glsl_type)

    def bind_to(self, target):
        """ Binds all inputs of this UBO to the given target, which may be
        either a RenderTarget or a NodePath """

        target.set_shader_inputs(**self._inputs)

    def update_input(self, name, value):
        """ Updates an existing input """
        self.ptas[name][0] = value

    def get_input(self, name):
        """ Returns the value of an existing input """
        return self.ptas[name][0]

    def generate_shader_code(self):  # pylint: disable=too-many-branches
        """ Generates the GLSL shader code to use the UBO """

        content = "#pragma once\n\n"
        content += "// Autogenerated by the render pipeline\n"
        content += "// Do not edit! Your changes will be lost.\n\n"

        structs = {}
        inputs = []

        for input_name, handle in iteritems(self.ptas):
            parts = input_name.split(".")

            # Single input, simply add it to the input list
            if len(parts) == 1:
                inputs.append(self.pta_to_glsl_type(handle) + " " + input_name + ";")

            # Nested input, like scattering.sun_color
            elif len(parts) == 2:
                struct_name = parts[0]
                actual_input_name = parts[1]
                if struct_name in structs:
                    # Struct is already defined, add member definition
                    structs[struct_name].append(self.pta_to_glsl_type(handle) + " " + actual_input_name + ";")
                else:
                    # Construct a new struct and add it to the list of inputs
                    inputs.append(struct_name + "_UBOSTRUCT " + struct_name + ";")
                    structs[struct_name] = [self.pta_to_glsl_type(handle) + " " + actual_input_name + ";"]

            # Nested input, like scattering.some_setting.sun_color, not supported yet
            else:
                self.warn("Structure definition too nested, not supported (yet):", input_name)

        # Add structures
        for struct_name, members in iteritems(structs):
            content += "struct " + struct_name + "_UBOSTRUCT {\n"
            for member in members:
                content += " " * 4 + member + "\n"
            content += "};\n\n"

        # Add actual inputs
        if len(inputs) < 1:
            self.debug("No UBO inputs present for", self.name)
        else:
            if self.use_ubo:

                content += "layout(shared, binding={}) uniform {}_UBO {{\n".format(self.bind_id, self.name)
                for ipt in inputs:
                    content += " " * 4 + ipt + "\n"
                content += "} " + self.name + ";\n"
            else:
                content += "uniform struct {\n"
                for ipt in inputs:
                    content += " " * 4 + ipt + "\n"
                content += "} " + self.name + ";\n"

        content += "\n"
        return content



import copy

from metadrive.component.map.scenario_map import ScenarioMap
from metadrive.constants import DEFAULT_AGENT
from metadrive.manager.base_manager import BaseManager
from metadrive.scenario.parse_object_state import parse_full_trajectory, parse_object_state, get_idm_route


class ScenarioMapManager(BaseManager):
    PRIORITY = 0  # Map update has the most high priority
    DEFAULT_DATA_BUFFER_SIZE = 200

    def __init__(self):
        super(ScenarioMapManager, self).__init__()
        self.store_map = self.engine.global_config.get("store_map", False)
        self.current_map = None
        self._no_map = self.engine.global_config["no_map"]
        self.map_num = self.engine.global_config["num_scenarios"]
        self.start_scenario_index = self.engine.global_config["start_scenario_index"]
        self._stored_maps = {
            i: None
            for i in range(self.start_scenario_index, self.start_scenario_index + self.map_num)
        }

        # we put the route searching function here
        self.sdc_start_point = None
        self.sdc_destinations = []
        self.sdc_dest_point = None
        self.current_sdc_route = None

    def reset(self):
        if not self._no_map:
            seed = self.engine.global_random_seed
            assert self.start_scenario_index <= seed < self.start_scenario_index + self.map_num

            self.current_sdc_route = None
            self.sdc_dest_point = None

            if self._stored_maps[seed] is None:
                new_map = ScenarioMap(map_index=seed)
                if self.store_map:
                    self._stored_maps[seed] = new_map
            else:
                new_map = self._stored_maps[seed]
            self.load_map(new_map)
        self.update_route()

    def update_route(self):
        data = self.engine.data_manager.current_scenario

        sdc_track = data.get_sdc_track()

        sdc_traj = parse_full_trajectory(sdc_track)

        init_state = parse_object_state(sdc_track, 0, check_last_state=False)
        last_state = parse_object_state(sdc_track, -1, check_last_state=True)
        init_position = init_state["position"]
        init_yaw = init_state["heading"]
        last_position = last_state["position"]
        last_yaw = last_state["heading"]

        self.current_sdc_route = get_idm_route(sdc_traj)
        self.sdc_start_point = copy.deepcopy(init_position)
        self.sdc_dest_point = copy.deepcopy(last_position)

        self.engine.global_config.update(
            copy.deepcopy(
                dict(
                    target_vehicle_configs={
                        DEFAULT_AGENT: dict(
                            spawn_position_heading=(init_position, init_yaw), spawn_velocity=init_state["velocity"]
                        )
                    }
                )
            )
        )

    def filter_path(self, start_lanes, end_lanes):
        for start in start_lanes:
            for end in end_lanes:
                path = self.current_map.road_network.shortest_path(start[0].index, end[0].index)
                if len(path) > 0:
                    return (start[0].index, end[0].index)
        return None

    def spawn_object(self, object_class, *args, **kwargs):
        raise ValueError("Please create ScenarioMap instance directly without calling spawn_object function.")
        # map = self.engine.spawn_object(object_class, auto_fill_random_seed=False, *args, **kwargs)
        # self.spawned_objects[map.id] = map
        # return map

    def load_map(self, map):
        map.attach_to_world()
        self.current_map = map

    def unload_map(self, map):
        map.detach_from_world()
        self.current_map = None
        if not self.engine.global_config["store_map"]:
            map.destroy()
            assert len(self.spawned_objects) == 0

    def destroy(self):
        self.maps = None
        self.current_map = None

        self.sdc_start_point = None
        self.sdc_destinations = []
        self.sdc_dest_point = None
        self.current_sdc_route = None

        super(ScenarioMapManager, self).destroy()

    def before_reset(self):
        # remove map from world before adding
        if self.current_map is not None:
            self.unload_map(self.current_map)

        self.sdc_start_point = None
        self.sdc_destinations = []
        self.sdc_dest_point = None
        self.current_sdc_route = None
        self.current_map = None

    def clear_objects(self, *args, **kwargs):
        """
        As Map instance should not be recycled, we will forcefully destroy useless map instances.
        """
        return super(ScenarioMapManager, self).clear_objects(force_destroy=True, *args, **kwargs)

    def clear_stored_maps(self):
        for m in self._stored_maps.values():
            if m is not None:
                m.detach_from_world()
                m.destroy()
        self._stored_maps = {
            i: None
            for i in range(self.start_scenario_index, self.start_scenario_index + self.map_num)
        }

    @property
    def num_stored_maps(self):
        return sum([1 if m is not None else 0 for m in self.engine.map_manager._stored_maps.values()])



from collections import OrderedDict


class Event:
    """
    Happening in episode level Event is a process for manipulating existing objects.
    All events will be cleared at the start of the episode
    A event/condition func should take its manager instance and other args as input
    """
    def __init__(
        self, name, condition_func, event_func, cond_args=None, event_args=None, trigger_once=True, priority=100
    ):
        assert DeprecationWarning("Not tested yet")
        self.name = name
        self.event_func = event_func
        self.event_args = event_args or {}
        self.condition_func = condition_func
        self.cond_args = cond_args or {}
        self.has_been_triggered = False
        self.trigger_once = trigger_once
        self.priority = priority

    def trigger(self):
        assert not self.has_been_triggered, "evnt can not be trigged twiced when trigger_once == True!"
        self.event_func(**self.event_args)
        self.has_been_triggered = True if self.trigger_once else False

    def should_be_triggered(self):
        return True if self.condition_func(**self.cond_args) else False


class EventManager:
    def __init__(self):
        assert DeprecationWarning("Not tested yet")
        self.events = OrderedDict()

        def _check_and_trigger_events(self):
            """
            This should usually be called in before_step()
            """
            for event in events.items():
                if event.should_be_triggered():
                    event.trigger()

        def get_untriggered_events(self):
            ret = []
            for name, event in self.events.items():
                if not event.has_been_triggered:
                    ret.append(name)
            return ret

        def get_triggered_events(self):
            ret = []
            for name, event in self.events.items():
                if event.trigger:
                    ret.append(name)
            return ret

        def _add_event(self, event):
            assert isinstance(event, Event), "Only event type can be added"
            self.events[event.name] = event
            self.events = OrderedDict(sorted(self.events.items(), key=lambda key, value: value.priority))

        def _delete_event(self, event_name):
            assert event_name in self.events, "No event in this manager {}".format(self)
            self.events.pop(event_name)



import copy
import numpy as np

from metadrive.component.traffic_light.scenario_traffic_light import ScenarioTrafficLight
from metadrive.type import MetaDriveType
from metadrive.scenario.scenario_description import ScenarioDescription as SD
from metadrive.manager.base_manager import BaseManager
import logging

logger = logging.getLogger(__file__)


class ScenarioLightManager(BaseManager):
    CLEAR_LIGHTS = False

    def __init__(self):
        super(ScenarioLightManager, self).__init__()
        self._scenario_id_to_obj_id = {}
        self._obj_id_to_scenario_id = {}
        self._lane_index_to_obj = {}
        self.skip_missing_light = self.engine.global_config["skip_missing_light"]
        self._episode_light_data = None

    def before_reset(self):
        super(ScenarioLightManager, self).before_reset()
        self._scenario_id_to_obj_id = {}
        self._lane_index_to_obj = {}
        self._obj_id_to_scenario_id = {}
        self._episode_light_data = self._get_episode_light_data()

    def after_reset(self):
        for scenario_lane_id, light_info in self._episode_light_data.items():
            if str(scenario_lane_id) not in self.engine.current_map.road_network.graph:
                logger.warning("Can not find lane for this traffic light. Skip!")
                if self.skip_missing_light:
                    continue
                else:
                    raise ValueError(
                        "Can not find lane for this traffic light. "
                        "Set skip_missing_light=True for skipping missing light!"
                    )
            lane_info = self.engine.current_map.road_network.graph[str(scenario_lane_id)]
            position = self._get_light_position(light_info)
            name = scenario_lane_id if self.engine.global_config["force_reuse_object_name"] else None
            traffic_light = self.spawn_object(ScenarioTrafficLight, lane=lane_info.lane, position=position, name=name)
            self._scenario_id_to_obj_id[scenario_lane_id] = traffic_light.id
            self._obj_id_to_scenario_id[traffic_light.id] = scenario_lane_id
            if self.engine.global_config["force_reuse_object_name"]:
                assert scenario_lane_id == traffic_light.id, "Original id should be assigned to traffic lights"
            self._lane_index_to_obj[lane_info.lane.index] = traffic_light
            status = light_info[SD.TRAFFIC_LIGHT_STATUS][self.episode_step]
            traffic_light.set_status(status)

    def _get_light_position(self, light_info):
        if SD.TRAFFIC_LIGHT_POSITION in light_info:
            # New format where the position is a 3-dim vector.
            return light_info[SD.TRAFFIC_LIGHT_POSITION]

        else:
            index = np.where(light_info[SD.TRAFFIC_LIGHT_LANE] > 0)[0][0]
            return light_info[SD.TRAFFIC_LIGHT_POSITION][index]

    def after_step(self, *args, **kwargs):
        if self.episode_step >= self.current_scenario_length:
            return

        for scenario_light_id, light_id, in self._scenario_id_to_obj_id.items():
            light_obj = self.spawned_objects[light_id]
            status = self._episode_light_data[scenario_light_id][SD.TRAFFIC_LIGHT_STATUS][self.episode_step]
            light_obj.set_status(status)

    def has_traffic_light(self, lane_index):
        return True if lane_index in self._lane_index_to_obj else False

    @property
    def current_scenario(self):
        return self.engine.data_manager.current_scenario

    @property
    def current_scenario_length(self):
        return self.engine.data_manager.current_scenario_length

    def _get_episode_light_data(self):
        ret = dict()
        for lane_id, light_info in self.current_scenario[SD.DYNAMIC_MAP_STATES].items():
            ret[lane_id] = copy.deepcopy(light_info[SD.STATE])
            ret[lane_id]["metadata"] = copy.deepcopy(light_info[SD.METADATA])

            if SD.TRAFFIC_LIGHT_POSITION in ret[lane_id]:
                # Old data format where position is a 2D array with shape [T, 2]
                traffic_light_position = ret[lane_id][SD.TRAFFIC_LIGHT_POSITION]

                if not np.any(ret[lane_id][SD.TRAFFIC_LIGHT_LANE].astype(bool)):
                    # This traffic light has no effect.
                    first_pos = -1
                else:
                    first_pos = np.argwhere(ret[lane_id][SD.TRAFFIC_LIGHT_LANE] != 0)[0, 0]
                traffic_light_position = traffic_light_position[first_pos]
            else:
                # New data format where position is a [3, ] array.
                traffic_light_position = light_info[SD.TRAFFIC_LIGHT_POSITION][:2]

            ret[lane_id][SD.TRAFFIC_LIGHT_POSITION] = traffic_light_position

            assert light_info[SD.TYPE] == MetaDriveType.TRAFFIC_LIGHT, "Can not handle {}".format(light_info[SD.TYPE])
        return ret

    def get_state(self):
        return {
            SD.OBJ_ID_TO_ORIGINAL_ID: copy.deepcopy(self._obj_id_to_scenario_id),
            SD.ORIGINAL_ID_TO_OBJ_ID: copy.deepcopy(self._scenario_id_to_obj_id)
        }



from metadrive.envs.metadrive_env import MetaDriveEnv

if __name__ == "__main__":
    env = MetaDriveEnv({
        "num_scenarios": 100,
        "start_seed": 5000,
        "traffic_density": 0.08,
    })
    env.reset()
    count = []
    for i in range(1, 101):
        o, r, tm, tc, info = env.step([0, 1])
        env.reset()
        # print(
        #     "Current map {}, vehicle number {}.".format(env.current_seed, env.engine.traffic_manager.get_vehicle_num())
        # )
        count.append(env.engine.traffic_manager.get_vehicle_num())
    # print(min(count), sum(count) / len(count), max(count))
    env.close()



from metadrive.engine.asset_loader import AssetLoader

try:
    from metadrive.envs.real_data_envs.nuplan_env import NuPlanEnv
except ImportError:
    pass
from metadrive.scenario.utils import draw_map
from metadrive.envs.metadrive_env import MetaDriveEnv
from metadrive.envs.real_data_envs.waymo_env import WaymoEnv
from metadrive.policy.idm_policy import WaymoIDMPolicy


def test_export_waymo_map(render=False):
    WaymoIDMPolicy.NORMAL_SPEED = 30
    asset_path = AssetLoader.asset_path
    env = WaymoEnv(
        {
            "manual_control": False,
            "no_traffic": True,
            "use_render": False,
            "data_directory": AssetLoader.file_path(asset_path, "waymo", return_raw_style=False),
            "num_scenarios": 3
        }
    )
    try:
        for seed in range(3):
            env.reset(seed=seed)
            map_vector = env.current_map.get_map_features()
            draw_map(map_vector, True if render else False)
    finally:
        env.close()


def test_metadrive_map_export(render=False):
    env = MetaDriveEnv(dict(image_observation=False, map=6, num_scenarios=1, start_seed=0))
    try:
        env.reset(seed=0)
        map_vector = env.current_map.get_map_features()
        draw_map(map_vector, True if render else False)
    finally:
        env.close()


def _test_nuplan_map_export(render=False):
    env = NuPlanEnv(
        {
            "DATASET_PARAMS": [
                'scenario_builder=nuplan_mini',
                'scenario_filter=one_continuous_log',  # simulate only one log
                "scenario_filter.log_names=['2021.09.16.15.12.03_veh-42_01037_01434']",
                'scenario_filter.limit_total_scenarios=1000',  # use 2 total scenarios
            ]
        }
    )
    try:
        env.reset(seed=0)
        map_vector = env.current_map.get_map_features()
        draw_map(map_vector, True if render else False)
    finally:
        env.close()


if __name__ == "__main__":
    # test_export_waymo_map(True)
    _test_nuplan_map_export(True)



import time

from metadrive.envs.metadrive_env import MetaDriveEnv

if __name__ == "__main__":
    env = MetaDriveEnv(
        {
            "num_scenarios": 1,
            "traffic_density": 0.1,
            "start_seed": 4,
            "vehicle_config": {
                # "stack_size": 5,
                "rgb_camera": (64, 64)
            },
            "manual_control": True,
            "use_render": False,
            "image_observation": True,  # it is a switch telling metadrive to use rgb as observation
            "rgb_clip": True,  # clip rgb to range(0,1) instead of (0, 255)
            # "pstats": True,
        }
    )
    env.reset()
    # # print m to capture rgb observation
    env.engine.accept(
        "m", env.vehicle.get_camera(env.vehicle.config["image_source"]).save_image, extraArgs=[env.vehicle]
    )
    start = time.time()
    for i in range(1, 100000):
        o, r, tm, tc, info = env.step([0, 0])
        assert env.observation_space.contains(o)
        # if i % 1000 == 0:
        # print("FPS: {}".format(i / (time.time() - start)))
        if tm or tc:
            # print("Reset")
            env.reset()
    env.close()



import gymnasium as gym
from metadrive.engine.engine_utils import get_global_config
import numpy as np

from metadrive.policy.base_policy import BasePolicy
from metadrive.utils.math import clip


class EnvInputPolicy(BasePolicy):
    DEBUG_MARK_COLOR = (252, 119, 3, 255)

    def __init__(self, obj, seed):
        # Since control object may change
        super(EnvInputPolicy, self).__init__(control_object=obj, random_seed=seed)
        self.discrete_action = self.engine.global_config["discrete_action"]
        self.use_multi_discrete = self.engine.global_config["use_multi_discrete"]
        self.steering_unit = 2.0 / (
            self.engine.global_config["discrete_steering_dim"] - 1
        )  # for discrete actions space
        self.throttle_unit = 2.0 / (
            self.engine.global_config["discrete_throttle_dim"] - 1
        )  # for discrete actions space
        self.discrete_steering_dim = self.engine.global_config["discrete_steering_dim"]
        self.discrete_throttle_dim = self.engine.global_config["discrete_throttle_dim"]

    def act(self, agent_id):
        action = self.engine.external_actions[agent_id]
        if self.engine.global_config["action_check"]:
            # Do action check for external input in EnvInputPolicy
            assert self.get_input_space().contains(action), "Input {} is not compatible with action space {}!".format(
                action, self.get_input_space()
            )
        if not self.discrete_action:
            to_process = action
        else:
            to_process = self.convert_to_continuous_action(action)

        # clip to -1, 1
        action = [clip(to_process[i], -1.0, 1.0) for i in range(len(to_process))]
        self.action_info["action"] = action
        return action

    def convert_to_continuous_action(self, action):
        # if isinstance(action, Iterable):
        #     assert len(action) == 2
        #     steering = action[0] * self.steering_unit - 1.0
        #     throttle = action[1] * self.throttle_unit - 1.0
        # else:
        # A more clear implementation:
        if self.use_multi_discrete:
            steering = action[0] * self.steering_unit - 1.0
            throttle = action[1] * self.throttle_unit - 1.0
        else:
            steering = float(action % self.discrete_steering_dim) * self.steering_unit - 1.0
            throttle = float(action // self.discrete_steering_dim) * self.throttle_unit - 1.0

        return steering, throttle

    @classmethod
    def get_input_space(cls):
        """
        The Input space is a class attribute
        """
        engine_global_config = get_global_config()
        extra_action_dim = engine_global_config["vehicle_config"]["extra_action_dim"]
        discrete_action = engine_global_config["discrete_action"]
        discrete_steering_dim = engine_global_config["discrete_steering_dim"]
        discrete_throttle_dim = engine_global_config["discrete_throttle_dim"]
        use_multi_discrete = engine_global_config["use_multi_discrete"]

        if not discrete_action:
            _input_space = gym.spaces.Box(-1.0, 1.0, shape=(2 + extra_action_dim, ), dtype=np.float32)
        else:
            if use_multi_discrete:
                _input_space = gym.spaces.MultiDiscrete([discrete_steering_dim, discrete_throttle_dim])
            else:
                _input_space = gym.spaces.Discrete(discrete_steering_dim * discrete_throttle_dim)
        return _input_space



import time
import cv2

import numpy as np
from metadrive import MetaDriveEnv


def _test_rgb_camera_as_obs(render=False):
    env = MetaDriveEnv(
        dict(
            num_scenarios=1,
            start_seed=1010,
            traffic_density=0.0,
            image_observation=True,
            use_render=False,
            vehicle_config=dict(image_source="rgb_camera", rgb_camera=(1200, 800)),
            show_interface=False,
            show_logo=False,
            show_fps=False,
        )
    )
    obs, _ = env.reset()
    action = [0.0, 0.1]
    start = time.time()
    for s in range(20000):
        o, r, tm, tc, i = env.step(action)
        # engine = env.engine
        # if engine.episode_step <= 1:
        #     engine.graphicsEngine.renderFrame()
        # origin_img = engine.win.getDisplayRegion(0).getScreenshot()
        # v = memoryview(origin_img.getRamImage).tolist()
        # img = np.array(v)
        # img = img.reshape((origin_img.getYSize(), origin_img.getXSize(), 4))
        # img = img[::-1]
        # img = img[..., :-1]
        # img = img/255
        if render:
            cv2.imshow("window", o["image"][..., -1])
            cv2.waitKey(1)
        #
        # if i % 100 == 0 and i != 0:
        # print("FPS: {}".format(i / (time.time() - start)))


if __name__ == "__main__":
    _test_rgb_camera_as_obs(True)



import logging

from metadrive.policy.base_policy import BasePolicy
from metadrive.scenario.parse_object_state import parse_object_state

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class ReplayTrafficParticipantPolicy(BasePolicy):
    """
       Replay policy from Real data. For adding new policy, overwrite get_trajectory_info()
       This policy is designed for Waymo Policy by default
       """
    DEBUG_MARK_COLOR = (3, 140, 252, 255)

    def __init__(self, control_object, track, random_seed=None):
        super(ReplayTrafficParticipantPolicy, self).__init__(control_object=control_object, random_seed=random_seed)
        self.start_index = 0
        self._velocity_local_frame = False
        self.traj_info = self.get_trajectory_info(track)

    @property
    def is_current_step_valid(self):
        return self.traj_info[self.episode_step] is not None

    def get_trajectory_info(self, track):
        ret = []
        for i in range(self.engine.data_manager.current_scenario_length):
            # a trick for saving computation
            if i < self.episode_step:
                ret.append(None)
            else:
                state = parse_object_state(track, i)
                if not state["valid"]:
                    ret.append(None)
                else:
                    ret.append(state)
        return ret

    def act(self, *args, **kwargs):
        index = max(int(self.episode_step), 0)
        if index >= len(self.traj_info):
            return None

        info = self.traj_info[index]

        # Before step
        # Warning by LQY: Don't call before step here! Before step should be called by manager
        # action = self.traj_info[int(self.episode_step)].get("action", None)
        # self.control_object.before_step(action)

        if not bool(info["valid"]):
            return None  # Return None action so the base vehicle will not overwrite the steering & throttle

        if "throttle_brake" in info:
            if hasattr(self.control_object, "set_throttle_brake"):
                self.control_object.set_throttle_brake(float(info["throttle_brake"]))
        if "steering" in info:
            if hasattr(self.control_object, "set_steering"):
                self.control_object.set_steering(float(info["steering"]))
        self.control_object.set_position(info["position"])
        self.control_object.set_velocity(info["velocity"], in_local_frame=self._velocity_local_frame)
        self.control_object.set_heading_theta(info["heading"])
        self.control_object.set_angular_velocity(info["angular_velocity"])

        return None  # Return None action so the base vehicle will not overwrite the steering & throttle


WaymoReplayTrafficParticipantPolicy = ReplayTrafficParticipantPolicy
ScenarioReplayTrafficParticipantPolicy = ReplayTrafficParticipantPolicy


class ReplayEgoCarPolicy(ReplayTrafficParticipantPolicy):
    def get_trajectory_info(self, trajectory):
        # Directly get trajectory from data manager
        trajectory_data = self.engine.data_manager.current_scenario["tracks"]
        sdc_track_index = str(self.engine.data_manager.current_scenario["metadata"]["sdc_id"])
        # if self.engine.data_manager.current_scenario["metadata"]["dataset"] == "nuplan":
        #     # nuplan local frame velocity
        #     self._velocity_local_frame = True
        ret = []
        for i in range(len(trajectory_data[sdc_track_index]["state"]["position"])):
            ret.append(parse_object_state(
                trajectory_data[sdc_track_index],
                i,
            ))
        return ret


WaymoReplayEgoCarPolicy = ReplayEgoCarPolicy
ScenarioReplayEgoCarPolicy = ReplayEgoCarPolicy


class NuPlanReplayEgoCarPolicy(ReplayEgoCarPolicy):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.sim_time_interval = self.engine.data_manager.time_interval
        if not self.control_object.config["no_wheel_friction"]:
            logger.warning("\nNOTE:set no_wheel_friction in vehicle config can make the replay more smooth! \n")

    def get_trajectory_info(self, *args, **kwargs):
        from metadrive.utils.nuplan.parse_object_state import parse_ego_vehicle_state_trajectory
        scenario = self.engine.data_manager.current_scenario
        return parse_ego_vehicle_state_trajectory(scenario, self.engine.current_map.nuplan_center)

    def act(self, *args, **kwargs):
        if self.episode_step >= len(self.traj_info):
            return

        self.control_object.set_position(self.traj_info[int(self.episode_step)]["position"])
        if self.episode_step < len(self.traj_info) - 1:
            velocity = self.traj_info[int(self.episode_step + 1)]["position"] - self.traj_info[int(self.episode_step
                                                                                                   )]["position"]
            velocity /= self.sim_time_interval
            self.control_object.set_velocity(velocity, in_local_frame=False)
        else:
            velocity = self.traj_info[int(self.episode_step)]["velocity"]
            self.control_object.set_velocity(velocity, in_local_frame=True)
        # self.control_object.set_velocity(self.traj_info[int(self.episode_step)]["velocity"])

        this_heading = self.traj_info[int(self.episode_step)]["heading"]
        angular_v = self.traj_info[int(self.episode_step)]["angular_velocity"]
        self.control_object.set_heading_theta(this_heading)
        self.control_object.set_angular_velocity(angular_v)

        return [0, 0]


class NuPlanReplayTrafficParticipantPolicy(BasePolicy):
    """
    This policy should be used with TrafficParticipantManager Together
    """
    def __init__(self, control_object, fix_height=None, random_seed=None, config=None):
        super(NuPlanReplayTrafficParticipantPolicy, self).__init__(control_object, random_seed, config)
        self.fix_height = fix_height

    def act(self, obj_state, *args, **kwargs):
        self.control_object.set_position(obj_state["position"], self.fix_height)
        self.control_object.set_heading_theta(obj_state["heading"])
        self.control_object.set_velocity(obj_state["velocity"])
        return [0, 0]



from typing import Optional, Union

from direct.showbase import OnScreenDebug
from panda3d.core import Vec4

from metadrive.constants import HELP_MESSAGE, DEBUG_MESSAGE


class ScreenMessage(OnScreenDebug.OnScreenDebug):
    """
    Simply inherit from the original debug class of panda3d to show debug message on screen
    """
    POS = (0.1, -0.2)
    SCALE = None

    def __init__(self, refresh_plain_text=False, debug=False):
        super(ScreenMessage, self).__init__()
        self.debug = debug
        self.enabled = True
        self.load()
        self.plain_text = set()
        self._refresh_plain_text = refresh_plain_text
        self._show_help_message = False

    def _update_data(self, data: Optional[Union[dict, str]]):
        # Several Node will be destructing or constructing again and again when debug pgraph
        self.onScreenText.cleanup()
        if isinstance(data, str):
            self.clear_all_plain_text()
            self.plain_text.add(data)
        elif isinstance(data, dict):
            for k, v in data.items():
                self.add(k, v)

    def load(self):
        super(ScreenMessage, self).load()
        self.onScreenText.setBg(Vec4(0, 0, 0, 0.5))
        self.onScreenText.setPos(*self.POS)
        self.onScreenText.textNode.setCardAsMargin(0.6, 0.6, 0.5, 0.1)
        if self.SCALE is not None:
            self.onScreenText.setScale(self.SCALE)

    def set_scale(self, scale=None):
        self.SCALE = scale

    def render(self, data: Optional[Union[dict, str]] = None):
        self._update_data(data)
        if not self.enabled:
            return
        if not self.onScreenText:
            self.load()
        self.onScreenText.clearText()
        if self._show_help_message:
            hlp_msg = HELP_MESSAGE + DEBUG_MESSAGE if self.debug else HELP_MESSAGE
            self.onScreenText.appendText(hlp_msg)
            return

        # Render plain text first
        for v in self.plain_text:
            self.onScreenText.appendText(v)

        # Render numerical values
        entries = list(self.data.items())
        entries.sort()
        for k, v in entries:
            if v[0] == self.frame:
                # It was updated this frame (key equals value):
                # isNew = " is"
                isNew = ":"
            else:
                # This data is not for the current
                # frame (key roughly equals value):
                # isNew = "was"
                isNew = ":"
            value = v[1]
            if type(value) == float:
                value = "% 10.4f" % (value, )
            # else: other types will be converted to str by the "%s"
            if type(value) == str:
                value = value.strip()
            if k:
                v_text = "%-100s\n" % (k.strip() + isNew + str(value))
            else:
                v_text = "{}\n".format(str(value))
            self.onScreenText.appendText(v_text)

        self.frame += 1

    def clear_plain_text(self, string):
        if string in self.plain_text:
            self.plain_text.remove(string)

    def clear_all_plain_text(self):
        self.plain_text.clear()

    def toggle_help_message(self):
        self.clear_all_plain_text()
        if self._show_help_message:
            self._show_help_message = False
        else:
            self._show_help_message = True



from metadrive.envs.metadrive_env import MetaDriveEnv


def local_test_close_and_restart():
    try:
        for m in ["X", "O", "C", "S", "R", "r", "T"]:
            env = MetaDriveEnv({"map": m, "use_render": True})
            o, _ = env.reset()
            for _ in range(300):
                assert env.observation_space.contains(o)
                o, r, tm, tc, i = env.step([1, 1])
                if tm or tc:
                    break
            env.close()
    finally:
        if "env" in locals():
            env = locals()["env"]
            env.close()


if __name__ == '__main__':
    local_test_close_and_restart()



import logging
import math
from typing import List, Dict

from panda3d.bullet import BulletBoxShape, BulletGhostNode
from panda3d.core import Vec3, LQuaternionf, Vec4, TextureStage, RigidBodyCombiner, \
    SamplerState, NodePath, Texture, Material

from metadrive.base_class.base_object import BaseObject
from metadrive.component.lane.abs_lane import AbstractLane
from metadrive.component.lane.point_lane import PointLane
from metadrive.component.road_network.node_road_network import NodeRoadNetwork
from metadrive.component.road_network.road import Road
from metadrive.constants import MetaDriveType, CamMask, PGLineType, PGLineColor, DrivableAreaProperty
from metadrive.engine.asset_loader import AssetLoader
from metadrive.engine.core.physics_world import PhysicsWorld
from metadrive.utils.coordinates_shift import panda_vector, panda_heading
from metadrive.utils.math import norm

logger = logging.getLogger(__name__)


class BaseBlock(BaseObject, DrivableAreaProperty):
    """
    Block is a driving area consisting of several roads
    Note: overriding the _sample() function to fill block_network/respawn_roads in subclass
    Call Block.construct_block() to add it to world
    """
    ID = "B"

    def __init__(
        self, block_index: int, global_network: NodeRoadNetwork, random_seed, ignore_intersection_checking=False
    ):
        super(BaseBlock, self).__init__(str(block_index) + self.ID, random_seed, escape_random_seed_assertion=True)
        # block information
        assert self.ID is not None, "Each Block must has its unique ID When define Block"
        assert len(self.ID) == 1, "Block ID must be a character "

        self.block_index = block_index
        self.ignore_intersection_checking = ignore_intersection_checking

        # each block contains its own road network and a global network
        self._global_network = global_network
        self.block_network = self.block_network_type()

        # a bounding box used to improve efficiency x_min, x_max, y_min, y_max
        self._bounding_box = None

        # used to spawn npc
        self._respawn_roads = []
        self._block_objects = None

        if self.render and not self.use_render_pipeline:
            self.ts_color = TextureStage("color")
            self.ts_normal = TextureStage("normal")
            self.ts_normal.setMode(TextureStage.M_normal)

            # Only maintain one copy of asset
            self.road_texture = self.loader.loadTexture(AssetLoader.file_path("textures", "sci", "new_color.png"))
            self.road_normal = self.loader.loadTexture(AssetLoader.file_path("textures", "sci", "normal.jpg"))
            self.road_texture.set_format(Texture.F_srgb)
            self.road_normal.set_format(Texture.F_srgb)
            self.road_texture.setMinfilter(SamplerState.FT_linear_mipmap_linear)
            self.road_texture.setAnisotropicDegree(8)

            # # continuous line
            # self.lane_line_model = self.loader.loadModel(AssetLoader.file_path("models", "box.bam"))
            # self.lane_line_model.setPos(0, 0, -DrivableAreaProperty.LANE_LINE_GHOST_HEIGHT / 2)
            self.lane_line_texture = self.loader.loadTexture(AssetLoader.file_path("textures", "sci", "floor.jpg"))
            # self.lane_line_model.setScale(DrivableAreaProperty.STRIPE_LENGTH*4,
            #                                    DrivableAreaProperty.LANE_LINE_WIDTH,
            #                                    DrivableAreaProperty.LANE_LINE_THICKNESS)
            # # self.lane_line_normal = self.loader.loadTexture(
            # #     AssetLoader.file_path("textures", "sci", "floor_normal.jpg"))
            # # self.lane_line_texture.set_format(Texture.F_srgb)
            # # self.lane_line_normal.set_format(Texture.F_srgb)
            # self.lane_line_model.setTexture(self.ts_color, self.lane_line_texture)
            # # self.lane_line_model.setTexture(self.ts_normal, self.lane_line_normal)
            #
            # # # broken line
            # self.broken_lane_line_model = self.loader.loadModel(AssetLoader.file_path("models", "box.bam"))
            # self.broken_lane_line_model.setScale(DrivableAreaProperty.STRIPE_LENGTH,
            #                                           DrivableAreaProperty.LANE_LINE_WIDTH,
            #                                           DrivableAreaProperty.LANE_LINE_THICKNESS)
            # self.broken_lane_line_model.setPos(0, 0, -DrivableAreaProperty.LANE_LINE_GHOST_HEIGHT / 2)
            # self.broken_lane_line_model.setTexture(self.ts_color, self.lane_line_texture)

            # side
            self.side_texture = self.loader.loadTexture(AssetLoader.file_path("textures", "sidewalk", "color.png"))
            self.side_texture.set_format(Texture.F_srgb)
            self.side_texture.setMinfilter(SamplerState.FT_linear_mipmap_linear)
            self.side_texture.setAnisotropicDegree(8)
            self.side_normal = self.loader.loadTexture(AssetLoader.file_path("textures", "sidewalk", "normal.png"))
            self.side_normal.set_format(Texture.F_srgb)
            self.sidewalk = self.loader.loadModel(AssetLoader.file_path("models", "box.bam"))
            self.sidewalk.setTwoSided(False)
            self.sidewalk.setTexture(self.ts_color, self.side_texture)
            # self.sidewalk = self.loader.loadModel(AssetLoader.file_path("models", "output.egg"))
            # self.sidewalk.setTexture(self.ts_normal, self.side_normal)

    def _sample_topology(self) -> bool:
        """
        Sample a new topology to fill self.block_network
        """
        raise NotImplementedError

    def construct_block(
        self,
        root_render_np: NodePath,
        physics_world: PhysicsWorld,
        extra_config: Dict = None,
        no_same_node=True,
        attach_to_world=True
    ) -> bool:
        """
        Randomly Construct a block, if overlap return False
        """
        self.sample_parameters()

        if not isinstance(self.origin, NodePath):
            self.origin = NodePath(self.name)
        # else:
        #     print("Origin already exists: ", self.origin)

        self._block_objects = []
        if extra_config:
            assert set(extra_config.keys()).issubset(self.PARAMETER_SPACE.parameters), \
                "Make sure the parameters' name are as same as what defined in pg_space.py"
            raw_config = self.get_config(copy=True)
            raw_config.update(extra_config)
            self.update_config(raw_config)
        self._clear_topology()
        success = self._sample_topology()
        self._global_network.add(self.block_network, no_same_node)

        self._create_in_world()
        self.attach_to_world(root_render_np, physics_world)

        if not attach_to_world:
            self.detach_from_world(physics_world)

        return success

    def destruct_block(self, physics_world: PhysicsWorld):
        self._clear_topology()
        self.detach_from_world(physics_world)

        self.origin.removeNode()
        self.origin = None

        self.dynamic_nodes.clear()
        self.static_nodes.clear()
        for obj in self._block_objects:
            obj.destroy()
        self._block_objects = None

    def construct_from_config(self, config: Dict, root_render_np: NodePath, physics_world: PhysicsWorld):
        success = self.construct_block(root_render_np, physics_world, config)
        return success

    def get_respawn_roads(self):
        return self._respawn_roads

    def get_respawn_lanes(self):
        """
        return a 2-dim array [[]] to keep the lane index
        """
        ret = []
        for road in self._respawn_roads:
            lanes = road.get_lanes(self.block_network)
            ret.append(lanes)
        return ret

    def get_intermediate_spawn_lanes(self):
        """Return all lanes that can be used to generate spawn intermediate vehicles."""
        raise NotImplementedError()

    def _add_one_respawn_road(self, respawn_road: Road):
        assert isinstance(respawn_road, Road), "Spawn roads list only accept Road Type"
        self._respawn_roads.append(respawn_road)

    def _clear_topology(self):
        if len(self._global_network.graph.keys()) > 0:
            self._global_network -= self.block_network
        self.block_network.graph.clear()
        self.PART_IDX = 0
        self.ROAD_IDX = 0
        self._respawn_roads.clear()

    """------------------------------------- For Render and Physics Calculation ---------------------------------- """

    def _create_in_world(self, skip=False):
        """
        Create NodePath and Geom node to perform both collision detection and render

        Note: Override the create_in_world() function instead of this one, since this method severing as a wrapper to
        help improve efficiency
        """
        self.lane_line_node_path = NodePath(RigidBodyCombiner(self.name + "_lane_line"))
        self.sidewalk_node_path = NodePath(RigidBodyCombiner(self.name + "_sidewalk"))
        self.lane_node_path = NodePath(RigidBodyCombiner(self.name + "_lane"))
        self.lane_vis_node_path = NodePath(RigidBodyCombiner(self.name + "_lane_vis"))

        if skip:  # for debug
            pass
        else:
            self.create_in_world()

        self.lane_line_node_path.flattenStrong()
        self.lane_line_node_path.node().collect()

        self.sidewalk_node_path.flattenStrong()
        self.sidewalk_node_path.node().collect()
        self.sidewalk_node_path.hide(CamMask.ScreenshotCam)

        # only bodies reparent to this node
        self.lane_node_path.flattenStrong()
        self.lane_node_path.node().collect()

        self.lane_vis_node_path.flattenStrong()
        self.lane_vis_node_path.node().collect()
        self.lane_vis_node_path.hide(CamMask.DepthCam | CamMask.ScreenshotCam)

        self.origin.hide(CamMask.Shadow)

        self.sidewalk_node_path.reparentTo(self.origin)
        self.lane_line_node_path.reparentTo(self.origin)
        self.lane_node_path.reparentTo(self.origin)
        self.lane_vis_node_path.reparentTo(self.origin)
        try:
            self._bounding_box = self.block_network.get_bounding_box()
        except:
            if len(self.block_network.graph) > 0:
                logging.warning("Can not find bounding box for it")
            self._bounding_box = None, None, None, None

        self._node_path_list.append(self.sidewalk_node_path)
        self._node_path_list.append(self.lane_line_node_path)
        self._node_path_list.append(self.lane_node_path)
        self._node_path_list.append(self.lane_vis_node_path)

    def create_in_world(self):
        """
        Create lane in the panda3D world
        """
        raise NotImplementedError

    def add_body(self, physics_body):
        raise DeprecationWarning(
            "Different from common objects like vehicle/traffic sign, Block has several bodies!"
            "Therefore, you should create BulletBody and then add them to self.dynamics_nodes "
            "manually. See in construct() method"
        )

    def get_state(self) -> Dict:
        """
        The record of Block type is not same as other objects
        """
        return {}

    def set_state(self, state: Dict):
        """
        Block type can not set state currently
        """
        pass

    def _add_lane_line(self, lane: AbstractLane, colors: List[Vec4], contruct_two_side=True):
        raise DeprecationWarning("Leave for argoverse using")
        if isinstance(lane, PointLane):
            parent_np = self.lane_line_node_path
            lane_width = lane.width_at(0)
            for c, i in enumerate([-1, 1]):
                line_color = colors[c]
                acc_length = 0
                if lane.line_types[c] == PGLineType.CONTINUOUS:
                    for segment in lane.segment_property:
                        lane_start = lane.position(acc_length, i * lane_width / 2)
                        acc_length += segment["length"]
                        lane_end = lane.position(acc_length, i * lane_width / 2)
                        middle = (lane_start + lane_end) / 2
                        self._add_lane_line2bullet(
                            lane_start, lane_end, middle, parent_np, line_color, lane.line_types[c]
                        )

    def _add_box_body(self, lane_start, lane_end, middle, parent_np: NodePath, line_type, line_color):
        raise DeprecationWarning("Useless, currently")
        length = norm(lane_end[0] - lane_start[0], lane_end[1] - lane_start[1])
        if PGLineType.prohibit(line_type):
            node_name = MetaDriveType.LINE_SOLID_SINGLE_WHITE if line_color == PGLineColor.GREY else MetaDriveType.LINE_SOLID_SINGLE_YELLOW
        else:
            node_name = MetaDriveType.BROKEN_LINE
        body_node = BulletGhostNode(node_name)
        body_node.set_active(False)
        body_node.setKinematic(False)
        body_node.setStatic(True)
        body_np = parent_np.attachNewNode(body_node)

        self._node_path_list.append(body_np)

        shape = BulletBoxShape(
            Vec3(length / 2, DrivableAreaProperty.LANE_LINE_WIDTH / 2, DrivableAreaProperty.LANE_LINE_GHOST_HEIGHT)
        )
        body_np.node().addShape(shape)
        mask = DrivableAreaProperty.CONTINUOUS_COLLISION_MASK if line_type != PGLineType.BROKEN else DrivableAreaProperty.BROKEN_COLLISION_MASK
        body_np.node().setIntoCollideMask(mask)
        self.static_nodes.append(body_np.node())

        body_np.setPos(panda_vector(middle, DrivableAreaProperty.LANE_LINE_GHOST_HEIGHT / 2))
        direction_v = lane_end - lane_start
        # theta = -numpy.arctan2(direction_v[1], direction_v[0])
        theta = panda_heading(math.atan2(direction_v[1], direction_v[0]))

        body_np.setQuat(LQuaternionf(math.cos(theta / 2), 0, 0, math.sin(theta / 2)))

    @property
    def block_network_type(self):
        """
        There are two type of road network to describe the relation of all lanes, override this func to assign one when
        you are building your own block.
        return: roadnetwork
        """
        raise NotImplementedError

    def destroy(self):
        if self.block_network is not None:
            self.block_network.destroy()
            if self.block_network.graph is not None:
                self.block_network.graph.clear()
            self.block_network = None
        self.PART_IDX = 0
        self.ROAD_IDX = 0
        self._respawn_roads.clear()
        self._global_network = None
        super(BaseBlock, self).destroy()

    def __del__(self):
        self.destroy()
        logger.debug("{} is being deleted.".format(type(self)))

    @property
    def bounding_box(self):
        return self._bounding_box



# Author: Epihaius
# Date: 2023-04-29

import array

from direct.showbase.ShowBase import ShowBase
from panda3d.core import *


class MyApp(ShowBase):
    def __init__(self):

        ShowBase.__init__(self)

        # set up a light source
        p_light = PointLight("point_light")
        p_light.set_color((1., 1., 1., 1.))
        self.light = self.camera.attach_new_node(p_light)
        self.light.set_pos(5., -10., 7.)
        self.render.set_light(self.light)
        self.disable_mouse()
        self.camera.set_pos(0., -50., 50.)
        self.camera.look_at(-10., 0., 0.)

        triangulator = Triangulator()
        values = array.array("f", [])
        vertex_data = GeomVertexData("poly", GeomVertexFormat.get_v3n3(), Geom.UH_static)

        with open("poly_coords.txt", "r") as coords_file:
            coords = coords_file.readlines()

        for i, coord in enumerate(coords):
            x, y = [float(c) for c in coord.replace("\n", "").split(" ")]
            values.extend((x, y, 0., 0., 0., 1.))
            triangulator.add_vertex(x, y)
            triangulator.add_polygon_vertex(i)

        triangulator.triangulate()
        prim = GeomTriangles(Geom.UH_static)

        for i in range(triangulator.get_num_triangles()):
            index0 = triangulator.get_triangle_v0(i)
            index1 = triangulator.get_triangle_v1(i)
            index2 = triangulator.get_triangle_v2(i)
            prim.add_vertices(index0, index1, index2)

        # add the values to the vertex table using a memoryview;
        # since the size of a memoryview cannot change, the vertex data table
        # already needs to have the right amount of rows before creating
        # memoryviews from its array(s)
        vertex_data.unclean_set_num_rows(len(coords))
        # retrieve the data array for modification
        data_array = vertex_data.modify_array(0)
        memview = memoryview(data_array).cast("B").cast("f")
        memview[:] = values

        geom = Geom(vertex_data)
        geom.add_primitive(prim)
        node = GeomNode("poly_node")
        node.add_geom(geom)
        poly = self.render.attach_new_node(node)
        poly.set_render_mode_wireframe()


app = MyApp()
app.run()



from typing import Tuple, Sequence, Union

import math
import numpy as np

from metadrive.component.lane.pg_lane import PGLane
from metadrive.constants import PGLineType
from metadrive.utils.math import norm


class StraightLane(PGLane):
    """A lane going in straight line."""
    def __init__(
        self,
        start: Union[np.ndarray, Sequence[float]],
        end: Union[np.ndarray, Sequence[float]],
        width: float = PGLane.DEFAULT_WIDTH,
        line_types: Tuple[PGLineType, PGLineType] = (PGLineType.BROKEN, PGLineType.BROKEN),
        forbidden: bool = False,
        speed_limit: float = 1000,
        priority: int = 0
    ) -> None:
        """
        New straight lane.

        :param start: the lane starting position [m]
        :param end: the lane ending position [m]
        :param width: the lane width [m]
        :param line_types: the type of lines on both sides of the lane
        :param forbidden: is changing to this lane forbidden
        :param priority: priority level of the lane, for determining who has right of way
        """
        super(StraightLane, self).__init__()
        self.set_speed_limit(speed_limit)
        self.start = np.array(start)
        self.end = np.array(end)
        self.width = width
        self.line_types = line_types or [PGLineType.BROKEN, PGLineType.BROKEN]
        self.forbidden = forbidden
        self.priority = priority
        self.length = norm((self.end - self.start)[0], (self.end - self.start)[1])
        self.heading = math.atan2(self.end[1] - self.start[1], self.end[0] - self.start[0])
        self.direction = (self.end - self.start) / self.length
        self.direction_lateral = np.array([self.direction[1], -self.direction[0]])

    def update_properties(self):
        super(StraightLane, self).__init__()
        self.length = norm((self.end - self.start)[0], (self.end - self.start)[1])
        self.heading = math.atan2(self.end[1] - self.start[1], self.end[0] - self.start[0])
        self.direction = (self.end - self.start) / self.length
        self.direction_lateral = np.array([self.direction[1], -self.direction[0]])

    def position(self, longitudinal: float, lateral: float) -> np.ndarray:
        return self.start + longitudinal * self.direction + lateral * self.direction_lateral

    def heading_theta_at(self, longitudinal: float) -> float:
        return self.heading

    def width_at(self, longitudinal: float) -> float:
        return self.width

    def local_coordinates(self, position: Tuple[float, float]) -> Tuple[float, float]:
        delta_x = position[0] - self.start[0]
        delta_y = position[1] - self.start[1]
        longitudinal = delta_x * self.direction[0] + delta_y * self.direction[1]
        lateral = delta_x * self.direction_lateral[0] + delta_y * self.direction_lateral[1]
        return float(longitudinal), float(lateral)

    def reset_start_end(self, start: Union[np.ndarray, Sequence[float]], end: Union[np.ndarray, Sequence[float]]):
        super(StraightLane, self).__init__()
        self.start = start
        self.end = end
        self.update_properties()

    def construct_lane_in_block(self, block, lane_index):
        """
        Straight lane can be represented by one segment
        """
        middle = self.position(self.length / 2, 0)
        end = self.position(self.length, 0)
        direction_v = end - middle
        theta = math.atan2(direction_v[1], direction_v[0])
        width = self.width_at(0) + block.SIDEWALK_LINE_DIST * 2
        self.construct_lane_segment(block, middle, width, self.length, theta, lane_index)

    @property
    def polygon(self):
        if self._polygon is None:
            polygon = []
            longs = np.arange(0, self.length + self.POLYGON_SAMPLE_RATE, self.POLYGON_SAMPLE_RATE)
            for k, lateral in enumerate([+self.width_at(0) / 2, -self.width_at(0) / 2]):
                if k == 1:
                    longs = longs[::-1]
                for longitude in longs:
                    point = self.position(longitude, lateral)
                    polygon.append([point[0], point[1]])
                    # polygon.append([point[0], point[1], 0.])
            self._polygon = np.asarray(polygon)
        return self._polygon



from typing import Dict, Union

from metadrive.utils.config import Config


class Configurable:
    """
    Instances of this class will maintain a config system, which is protected from unexpected modification
    """
    def __init__(self, config: Union[Dict, Config] = None):
        # initialize and specify the value in config
        self._config = Config(config if config is not None else {})

    def get_config(self, copy=True) -> Union[Config, Dict]:
        """
        Return self._config
        :param copy:
        :return: a copy of config dict
        """
        if copy:
            return self._config.copy()
        return self._config

    def update_config(self, config: dict, allow_add_new_key=False):
        """
        Merge config and self._config
        """
        self._config.update(config, allow_add_new_key=allow_add_new_key)

    def destroy(self):
        """
        Fully delete this element and release the memory
        """
        self._config.clear()
        # if hasattr(self, "engine"):
        #     self.engine = None

    @property
    def config(self):
        return self._config



import time

import cv2

from metadrive import MetaDriveEnv
from metadrive.policy.idm_policy import IDMPolicy


def _test_depth_camera_as_obs(render=False):
    env = MetaDriveEnv(
        dict(
            num_scenarios=1,
            start_seed=1010,
            agent_policy=IDMPolicy,
            traffic_density=0.0,
            image_observation=True,
            image_on_cuda=True,
            use_render=False,
            vehicle_config=dict(image_source="depth_camera", depth_camera=(800, 600)),
            show_interface=True,
            show_logo=False,
            show_fps=False,
        )
    )
    obs, _ = env.reset()
    action = [0.0, 0.1]
    start = time.time()
    for i in range(20000):
        o, r, tm, tc, _ = env.step(action)
        if render:
            ret = o["image"].get()[..., -1] if env.config["image_on_cuda"] else o["image"][..., -1]
            cv2.imshow("window", ret)
            cv2.waitKey(1)
        # if d:
        # print("FPS: {}".format(i / (time.time() - start)))
        # env.reset()
        # break


def _test_main_rgb_camera_as_obs_with_interface(render=False):
    env = MetaDriveEnv(
        dict(
            num_scenarios=1,
            start_seed=1010,
            agent_policy=IDMPolicy,
            traffic_density=0.0,
            image_observation=True,
            image_on_cuda=True,
            use_render=False,
            vehicle_config=dict(image_source="main_camera", rgb_camera=(800, 600)),
            show_interface=True,
            show_logo=False,
            show_fps=False,
        )
    )
    obs, _ = env.reset()
    action = [0.0, 0.1]
    start = time.time()
    reset_time = 0
    for i in range(20000):
        o, r, tm, tc, _ = env.step(action)
        if render:
            ret = o["image"].get()[..., -1] if env.config["image_on_cuda"] else o["image"][..., -1]
            cv2.imshow("window", ret)
            cv2.waitKey(1)
        if tm or tc:
            current = time.time()
            # env.reset()
            # reset_time += time.time()-current
            # print("FPS: {}".format(i / (current - start - reset_time)))
            break


def _test_main_rgb_camera_no_interface(render=False):
    env = MetaDriveEnv(
        dict(
            num_scenarios=1,
            start_seed=1010,
            agent_policy=IDMPolicy,
            traffic_density=0.0,
            image_observation=True,
            image_on_cuda=True,
            use_render=False,
            vehicle_config=dict(image_source="main_camera", rgb_camera=(800, 600)),
            show_interface=False,
            show_logo=False,
            show_fps=False,
        )
    )
    obs, _ = env.reset()
    action = [0.0, 0.1]
    start = time.time()
    for i in range(20000):
        o, r, tm, tc, _ = env.step(action)
        if render:
            ret = o["image"].get()[..., -1] if env.config["image_on_cuda"] else o["image"][..., -1]
            cv2.imshow("window", ret)
            cv2.waitKey(1)
        if tm or tc:
            # print("FPS: {}".format(i / (time.time() - start)))
            # env.reset()
            break


def _test_rgb_camera_as_obs(render=False):
    env = MetaDriveEnv(
        dict(
            num_scenarios=1,
            start_seed=1010,
            agent_policy=IDMPolicy,
            traffic_density=0.0,
            image_observation=True,
            image_on_cuda=True,
            use_render=False,
            vehicle_config=dict(image_source="rgb_camera", rgb_camera=(1920, 1080)),
            show_interface=False,
            show_logo=False,
            show_fps=False,
        )
    )
    obs, _ = env.reset()
    action = [0.0, 0.1]
    start = time.time()
    for i in range(20000):
        o, r, tm, tc, _ = env.step(action)
        if render:
            ret = o["image"].get()[..., -1] if env.config["image_on_cuda"] else o["image"][..., -1]
            cv2.imshow("window", ret)
            cv2.waitKey(1)
        if tm or tc:
            # print("FPS: {}".format(i / (time.time() - start)))
            env.reset()
            # break


if __name__ == "__main__":
    # _test_depth_camera_as_obs(False)
    _test_rgb_camera_as_obs(True)
    # _test_main_rgb_camera_as_obs_with_interface(False)
    # _test_main_rgb_camera_no_interface(True)



# Copyright (c) 2021 kitsune.ONE team.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

bl_info = {
    'name': 'KITSUNETSUKI Asset Tools',
    'author': 'kitsune.ONE team',
    'version': (0, 6, 9),
    'blender': (2, 92, 0),
    'location': 'File > Import-Export',
    'description': 'Exports: glTF, VRM',
    'warning': '',
    'support': 'COMMUNITY',
    'wiki_url': '',
    'tracker_url': 'https://github.com/kitsune-ONE-team/KITSUNETSUKI-Asset-Tools/issues',
    'category': 'Import-Export',
}


def reload_package(module_dict_main):
    import importlib
    from pathlib import Path
    from typing import Any, Dict

    def reload_package_recursive(current_dir: Path, module_dict: Dict[str, Any]) -> None:
        for path in current_dir.iterdir():
            if '__init__' in str(path) or path.stem not in module_dict:
                continue

            if path.is_file() and path.suffix == '.py':
                importlib.reload(module_dict[path.stem])
            elif path.is_dir():
                reload_package_recursive(path, module_dict[path.stem].__dict__)

    reload_package_recursive(Path(__file__).parent, module_dict_main)


if 'bpy' in locals():
    reload_package(locals())


def register():
    import bpy

    if bpy.app.version[:2] < bl_info['blender'][:2]:
        cur_ver = '{}.{}'.format(*bpy.app.version[:2])
        req_ver = '{}.{}'.format(*bl_info['blender'][:2])
        raise Exception(
            f"This add-on doesn't support Blender version less than {req_ver}. "
            f'Blender version {req_ver} or greater is recommended, '
            f'but the current version is {cur_ver}'
        )

    from . import blend2gltf, blend2vrm
    blend2gltf.register(bl_info['version'])
    blend2vrm.register(bl_info['version'])


def unregister():
    import bpy

    if bpy.app.version < bl_info['blender'][:2]:
        return

    from . import blend2gltf, blend2vrm
    blend2gltf.unregister()
    blend2vrm.unregister()


if __name__ == '__main__':
    register()



import os

import cv2
from tqdm.auto import tqdm


def image_files_to_video(video_name, image_folder, code="mp4v"):
    """
    code=mp4v, avc1, x264, h264 etc.
    """
    assert video_name.endswith(".mp4")
    images = [img for img in os.listdir(image_folder) if img.endswith(".png")]
    images.sort(key=lambda x: int(x[:-4]))
    assert len(images) > 0
    frame = cv2.imread(os.path.join(image_folder, images[0]))
    height, width, layers = frame.shape
    video = cv2.VideoWriter(video_name, cv2.VideoWriter_fourcc(*code), 40, (width, height))
    for image in tqdm(images, desc="Writing Video"):
        video.write(cv2.imread(os.path.join(image_folder, image)))
    video.release()
    try:
        cv2.destroyAllWindows()
    except Exception as e:
        print("Error happen: ", e)


def image_list_to_video(video_name, image_list, code="mp4v"):
    """
    code=mp4v, avc1, x264, h264 etc.
    """
    assert video_name.endswith(".mp4")
    assert len(image_list) > 0
    # frame = cv2.imread(os.path.join(image_folder, images[0]))
    frame = image_list[0]
    height, width, layers = frame.shape
    video = cv2.VideoWriter(video_name, cv2.VideoWriter_fourcc(*code), 40, (width, height))
    for image in tqdm(image_list, desc="Writing Video"):

        # Change color
        image = image[..., ::-1]

        video.write(image)
    video.release()
    try:
        cv2.destroyAllWindows()
    except Exception as e:
        print("Error happen: ", e)



import time

from metadrive.engine.asset_loader import AssetLoader
from metadrive.envs.scenario_env import ScenarioEnv
from metadrive.policy.replay_policy import ReplayEgoCarPolicy

NuScenesEnv = ScenarioEnv

if __name__ == "__main__":
    env = NuScenesEnv(
        {
            "use_render": True,
            "no_map": False,
            "agent_policy": ReplayEgoCarPolicy,
            # "manual_control": True,
            "show_interface": False,
            # "need_lane_localization": False,
            # "image_observation": True,
            "show_logo": False,
            "no_traffic": False,
            "store_data": False,
            "sequential_seed": True,
            # "debug_static_world": True,
            # "sequential_seed": True,
            "reactive_traffic": True,
            "curriculum_level": 1,
            "show_fps": False,
            # "debug": True,
            "no_static_vehicles": False,
            # "pstats": True,
            "render_pipeline": True,
            # "daytime": "22:01",
            # "no_traffic": True,
            # "no_light": False,
            # "debug":True,
            # Make video
            # "episodes_to_evaluate_curriculum": 5,
            "window_size": (1600, 900),
            "camera_dist": 9,
            # "camera_height": 0.5,
            # "camera_pitch": np.pi / 3,
            # "camera_fov": 60,
            # "force_render_fps": 10,
            "start_scenario_index": 0,
            "num_scenarios": 10,
            # "force_reuse_object_name": True,
            # "data_directory": "/home/shady/Downloads/test_processed",
            "horizon": 1000,
            # "no_static_vehicles": True,
            # "show_policy_mark": True,
            # "show_coordinates": True,
            # "force_destroy": True,
            # "default_vehicle_in_traffic": True,
            "vehicle_config": dict(
                # light=True,
                # random_color=True,
                show_navi_mark=False,
                image_source="rgb_camera",
                rgb_camera=(1600, 900),
                # no_wheel_friction=True,
                lidar=dict(num_lasers=120, distance=50),
                lane_line_detector=dict(num_lasers=0, distance=50),
                side_detector=dict(num_lasers=12, distance=50)
            ),
            "data_directory": "/home/shady/data/scenarionet/dataset/nuscenes"
        }
    )

    # 0,1,3,4,5,6

    success = []
    reset_num = 0
    start = time.time()
    reset_used_time = 0
    s = 0
    while True:
        # for i in range(10):
        start_reset = time.time()
        env.reset(seed=0)

        reset_used_time += time.time() - start_reset
        reset_num += 1
        for t in range(10000):
            o, r, tm, tc, info = env.step([1, 0.88])
            assert env.observation_space.contains(o)
            s += 1
            if env.config["use_render"]:
                env.render(
                    text={
                        "seed": env.current_seed,
                        "num_map": info["num_stored_maps"],
                        "data_coverage": info["data_coverage"],
                        "reward": r,
                        "heading_r": info["step_reward_heading"],
                        "lateral_r": info["step_reward_lateral"],
                        "smooth_action_r": info["step_reward_action_smooth"]
                    },
                    # mode="topdown"
                )
            if tm or tc:
                print(
                    "Time elapse: {:.4f}. Average FPS: {:.4f}, AVG_Reset_time: {:.4f}".format(
                        time.time() - start, s / (time.time() - start - reset_used_time), reset_used_time / reset_num
                    )
                )
                print("seed:{}, success".format(env.engine.global_random_seed))
                print(list(env.engine.curriculum_manager.recent_success.dict.values()))
                break



import pickle

from metadrive.engine.asset_loader import AssetLoader
from metadrive.scenario.scenario_description import ScenarioDescription

import logging

logger = logging.getLogger(__name__)

logger.warning(
    "This converters will de deprecated. "
    "Use tools in ScenarioNet instead: https://github.com/metadriverse/ScenarioNet."
)

if __name__ == '__main__':

    with open("waymo120/0408_output_final/dataset_summary.pkl", "rb") as f:
        summary_dict = pickle.load(f)

    new_summary = {}
    for obj_id, summary in summary_dict.items():

        if summary["number_summary"]["dynamic_object_states"] == 0:
            continue

        if summary["object_summary"]["sdc"]["distance"] < 80 or \
                summary["object_summary"]["sdc"]["continuous_valid_length"] < 50:
            continue

        if len(summary["number_summary"]["object_types"]) < 3:
            continue

        if summary["number_summary"]["object"] < 80:
            continue

        new_summary[obj_id] = summary

        if len(new_summary) >= 3:
            break

    file_path = AssetLoader.file_path("waymo", ScenarioDescription.DATASET.SUMMARY_FILE, return_raw_style=False)
    with open(file_path, "wb") as f:
        pickle.dump(new_summary, f)

    print(new_summary.keys())



# Please don't change the order of following packages!
import os
import sys
from os import path

from setuptools import setup, find_namespace_packages  # This should be place at top!

ROOT_DIR = os.path.dirname(__file__)


def get_version():
    context = {}
    with open('./metadrive/version.py', 'r') as file:
        exec(file.read(), context)
    return context['VERSION']


VERSION = get_version()


def is_mac():
    return sys.platform == "darwin"


def is_win():
    return sys.platform == "win32"


assert sys.version_info.major == 3 and sys.version_info.minor >= 6 and sys.version_info.minor < 12, \
    "python version >= 3.6, <3.12 is required"

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()
packages = find_namespace_packages(
    exclude=("docs", "docs.*", "documentation", "documentation.*", "build.*"))
print("We will install the following packages: ", packages)

install_requires = [
    "requests",
    "gymnasium>=0.28, <0.29",
    "numpy>=1.21.6, <=1.24.2",
    "matplotlib",
    "pandas",
    "pygame",
    "tqdm",
    "yapf",
    "seaborn",
    "tqdm",
    "progressbar",
    # "panda3d==1.10.8",
    "panda3d==1.10.13",
    "panda3d-gltf==0.13",  # 0.14 will bring some problems
    "panda3d-simplepbr",
    "pillow",
    "protobuf==3.20.3",
    "pytest",
    "opencv-python",
    "lxml",
    "scipy",
    "psutil",
    "geopandas",
    "shapely"
]

nuplan_requirement = [
    "bokeh==2.4",
    "hydra-core",
    "chardet",
    "pyarrow",
    "aiofiles",
    "retry",
    "boto3",
    "aioboto3"
]

# Or try:
#   pip install git+https://github.com/waymo-research/waymo-open-dataset.git
waymo_requirement = [
    "waymo-open-dataset-tf-2.11.0==1.5.0"
]

cuda_requirement = [
    "cuda-python==12.0.0",
    "PyOpenGL==3.1.6",
    "PyOpenGL-accelerate==3.1.6",
    "pyrr==0.10.3",
    "glfw",
]

gym_requirement = [
    "gym>=0.19.0, <=0.26.0"
]

setup(
    name="metadrive-simulator",
    python_requires='>=3.6, <3.12',  # do version check with assert
    version=VERSION,
    description="An open-ended driving simulator with infinite scenes",
    url="https://github.com/metadriverse/metadrive",
    author="MetaDrive Team",
    author_email="quanyili0057@gmail.com, pzh@cs.ucla.edu",
    packages=packages,
    install_requires=install_requires,
    extras_require={
        "cuda": cuda_requirement,
        "nuplan": nuplan_requirement,
        "waymo": waymo_requirement,
        "gym": gym_requirement,
        "all": nuplan_requirement + cuda_requirement
    },
    include_package_data=True,
    license="Apache 2.0",
    long_description=long_description,
    long_description_content_type='text/markdown',
)

"""
How to publish to pypi and Draft github Release?  Noted by Zhenghao and Quanyi in Dec 27, 2020.

0. Checkout a new branch from main called releases/x.y.z

1. Rename VERSION in metadrive/version.py to x.y.z

2. Revise the version in metadrive/assets/version.txt, and compress the folder into assets.zip

3. commit changes and push this branch to remote

4. Draft a release on github with new version number and upload assets.zip 

5. Remove old files and ext_modules from setup() to get a clean wheel for all platforms in py3-none-any.wheel
    rm -rf dist/ build/ documentation/build/ metadrive_simulator.egg-info/ docs/build/

6. Rename current version to X.Y.Z.rcA, where A is arbitrary value represent "release candidate A". 
   This is really important since pypi do not support renaming and re-uploading. 
   Rename version in metadrive/versions.py 

7. Get wheel
    python setup.py sdist bdist_wheel

8. Upload to test channel
    twine upload --repository testpypi dist/*

9. Test as next line. If failed, change the version name and repeat 1, 2, 3, 4, 5.
    pip install --index-url https://test.pypi.org/simple/ metadrive

10. Rename current version to X.Y.Z in setup.py, rerun 1, 3 steps.

11. Upload to production channel 
    twine upload dist/*

12. Upload the generated .whl file to release files

13. merge this branch into main

!!!!!!!!!!!!! NOTE: please make sure that unzip assets.zip will generate a folder called assets instead of files  

"""



from metadrive.envs.marl_envs.marl_inout_roundabout import MultiAgentRoundaboutEnv


def test_infinite_agents():
    env = MultiAgentRoundaboutEnv(
        {
            "map_config": dict(exit_length=20, lane_num=2),
            # "use_render": True,
            "num_agents": -1,
            "delay_done": 50,
            "horizon": 50,
            "debug": True,
            "random_traffic": False
        }
    )
    try:
        o, _ = env.reset()
        env.seed(100)
        env._DEBUG_RANDOM_SEED = 100
        max_num = old_num_of_vehicles = len(env.vehicles)
        for i in range(1, 1000):
            o, r, tm, tc, info = env.step({k: [1, 1] for k in env.vehicles})
            # # print(
            #     "{} Current active agents: ".format(i), len(env.vehicles), ". Objects: ",
            #     len(env.agent_manager._object_to_agent)
            # )
            max_num = max(len(env.vehicles), max_num)
            # env.render(mode="top_down")
            for kkk, iii in info.items():
                if tm[kkk] or tc[kkk]:
                    assert iii["episode_length"] >= 1
            if tm["__all__"]:
                o, _ = env.reset()
                # # print("Finish {} steps.".format(i))
    finally:
        env._DEBUG_RANDOM_SEED = None
        env.close()

    env = MultiAgentRoundaboutEnv({"num_agents": -1, "delay_done": 0, "horizon": 50, "debug": True})
    try:
        o, _ = env.reset()
        max_num = old_num_of_vehicles = len(env.vehicles)
        for i in range(1, 300):
            o, r, tm, tc, info = env.step({k: [0, 1] for k in env.vehicles})
            # # print("Current active agents: ", len(env.vehicles),
            #       ". Objects: ", len(env.agent_manager._object_to_agent))
            max_num = max(len(env.vehicles), max_num)
            # env.render(mode="top_down")
            for kkk, iii in info.items():
                if tm[kkk] or tc[kkk]:
                    assert iii["episode_length"] >= 1
            if tm["__all__"]:
                o, _ = env.reset()
                # print("Finish {} steps.".format(i))
    finally:
        env.close()
    assert max_num >= old_num_of_vehicles


if __name__ == '__main__':
    test_infinite_agents()



from metadrive.component.pgblock.intersection import InterSection
from metadrive.component.pgblock.pg_block import PGBlockSocket
from metadrive.component.road_network import Road
from metadrive.constants import Goal, PGLineType, PGLineColor
from metadrive.component.pg_space import ParameterSpace, Parameter, BlockParameterSpace


class TInterSection(InterSection):
    """
    A structure like X Intersection, code copied from it mostly
    """

    ID = "T"
    SOCKET_NUM = 2
    PARAMETER_SPACE = ParameterSpace(BlockParameterSpace.T_INTERSECTION)

    def _try_plug_into_previous_block(self) -> bool:
        no_cross = super(TInterSection, self)._try_plug_into_previous_block()
        self._exclude_lanes()
        return no_cross

    def _change_vis(self, t_type):
        # not good here,
        next_part_socket = self.get_socket_list()[(t_type + 1) % 4]
        # next_part_socket = self._sockets[(t_type + 1) % 4]  # FIXME pzh: Help! @LQY What is in this part?
        next_positive = next_part_socket.positive_road
        next_negative = next_part_socket.negative_road

        last_part_socket = self.get_socket_list()[(t_type + 3) % 4]
        # last_part_socket = self._sockets[(t_type + 3) % 4]  # FIXME pzh: Help! @LQY
        last_positive = last_part_socket.positive_road
        last_negative = last_part_socket.negative_road
        if t_type == Goal.LEFT:
            next_positive = next_part_socket.negative_road
            next_negative = next_part_socket.positive_road
        if t_type == Goal.RIGHT:
            last_positive = last_part_socket.negative_road
            last_negative = last_part_socket.positive_road

        for i, road in enumerate([Road(last_negative.end_node, next_positive.start_node),
                                  Road(next_negative.end_node, last_positive.start_node)]):
            lanes = road.get_lanes(self.block_network)
            outside_type = PGLineType.SIDE if i == 0 else PGLineType.NONE
            for k, lane in enumerate(lanes):
                line_types = [PGLineType.BROKEN, PGLineType.BROKEN
                              ] if k != len(lanes) - 1 else [PGLineType.BROKEN, outside_type]
                lane.line_types = line_types
                if k == 0:
                    lane.line_colors = [PGLineColor.YELLOW, PGLineColor.GREY]
                    if i == 1:
                        lane.line_types[0] = PGLineType.NONE

    def _exclude_lanes(self):
        para = self.get_config()
        t_type = para[Parameter.t_intersection_type]
        self.add_sockets(self.pre_block_socket)

        start_node = self._sockets[PGBlockSocket.get_real_index(self.name, t_type)].negative_road.end_node
        end_node = self._sockets[PGBlockSocket.get_real_index(self.name, t_type)].positive_road.start_node
        for i in range(4):
            if i == t_type:
                continue
            index_i = PGBlockSocket.get_real_index(self.name, i) if i < 3 else self.pre_block_socket_index
            exit_node = self._sockets[index_i].positive_road.start_node if i != Goal.ADVERSE else self._sockets[
                index_i].negative_road.start_node
            pos_lanes = self.block_network.remove_all_roads(start_node, exit_node)
            entry_node = self._sockets[index_i].negative_road.end_node if i != Goal.ADVERSE else self._sockets[
                index_i].positive_road.end_node
            neg_lanes = self.block_network.remove_all_roads(entry_node, end_node)

            # if (i + 2) % 4 == t_type:
            #     # for vis only, solve a bug existing in a corner case,
            #     for lane in neg_lanes:
            #         lane.reset_start_end(lane.start, lane.position(lane.length / 2, 0))
            #     self.block_network.add_road(Road(Decoration.start, Decoration.end), neg_lanes)
            #
            #     for lane in pos_lanes:
            #         lane.reset_start_end(lane.position(lane.length / 2, 0), lane.end)
            #     self.block_network.add_road(Road(Decoration.start, Decoration.end), pos_lanes)

        self._change_vis(t_type)
        self._sockets.pop(self.pre_block_socket.index)
        socket = self._sockets.pop(PGBlockSocket.get_real_index(self.name, t_type))
        self.block_network.remove_all_roads(socket.positive_road.start_node, socket.positive_road.end_node)
        self.block_network.remove_all_roads(socket.negative_road.start_node, socket.negative_road.end_node)
        self._respawn_roads.remove(socket.negative_road)

    def _add_one_socket(self, socket: PGBlockSocket):
        assert isinstance(socket, PGBlockSocket), "Socket list only accept BlockSocket Type"

        # mute warning in T interection

        # if socket.index is not None and not socket.index.startswith(self.name):
        #     logging.warning(
        #         "The adding socket has index {}, which is not started with this block name {}. This is dangerous! "
        #         "Current block has sockets: {}.".format(socket.index, self.name, self.get_socket_indices())
        #     )
        if socket.index is None:
            # if this socket is self block socket
            socket.set_index(self.name, len(self._sockets))
        self._sockets[socket.index] = socket

    def enable_u_turn(self, enable_u_turn: bool):
        raise ValueError("T intersection didn't support u_turn now")



from metadrive.engine.core.manual_controller import KeyboardController, SteeringWheelController, XboxController
from metadrive.engine.engine_utils import get_global_config
from metadrive.examples import expert
from metadrive.policy.env_input_policy import EnvInputPolicy


class ManualControlPolicy(EnvInputPolicy):
    """
    Control the current track vehicle
    """

    DEBUG_MARK_COLOR = (252, 244, 3, 255)

    def __init__(self, obj, seed, enable_expert=True):
        super(ManualControlPolicy, self).__init__(obj, seed)
        config = self.engine.global_config
        self.enable_expert = enable_expert

        if config["manual_control"] and config["use_render"]:
            self.engine.accept("t", self.toggle_takeover)
            pygame_control = False
        elif config["manual_control"]:
            # Use pygame to accept key strike.
            pygame_control = True
        else:
            pygame_control = False

        # if config["manual_control"] and config["use_render"]:
        if config["manual_control"]:
            if config["controller"] == "keyboard":
                self.controller = KeyboardController(pygame_control=pygame_control)
            elif config["controller"] in ["xboxController", "xboxcontroller", "xbox", "joystick"]:
                try:
                    if config["controller"] == "joystick":
                        self.controller = SteeringWheelController()
                    else:
                        self.controller = XboxController()
                except:
                    print("Load Joystick Error! Fall back to keyboard control")
                    self.controller = KeyboardController(pygame_control=pygame_control)
            else:
                raise ValueError("No such a controller type: {}".format(self.config["controller"]))
        else:
            self.controller = None

    def act(self, agent_id):

        self.controller.process_others(takeover_callback=self.toggle_takeover)

        try:
            if self.engine.current_track_vehicle.expert_takeover and self.enable_expert:
                return expert(self.engine.current_track_vehicle)
        except (ValueError, AssertionError):
            # if observation doesn't match, fall back to manual control
            print("Current observation does not match the format that expert can accept.")
            self.toggle_takeover()

        is_track_vehicle = self.engine.agent_manager.get_agent(agent_id) is self.engine.current_track_vehicle
        not_in_native_bev = (self.engine.main_camera is None) or (not self.engine.main_camera.is_bird_view_camera())
        if self.engine.global_config["manual_control"] and is_track_vehicle and not_in_native_bev:
            action = self.controller.process_input(self.engine.current_track_vehicle)
            self.action_info["manual_control"] = True
        else:
            action = super(ManualControlPolicy, self).act(agent_id)
            self.action_info["manual_control"] = False

        self.action_info["action"] = action
        return action

    def toggle_takeover(self):
        if self.engine.current_track_vehicle is not None:
            self.engine.current_track_vehicle.expert_takeover = not self.engine.current_track_vehicle.expert_takeover
            print("The expert takeover is set to: ", self.engine.current_track_vehicle.expert_takeover)


class TakeoverPolicy(EnvInputPolicy):
    """
    Record the takeover signal
    """
    def __init__(self, obj, seed):
        super(TakeoverPolicy, self).__init__(obj, seed)
        config = get_global_config()
        if config["manual_control"] and config["use_render"]:
            if config["controller"] == "joystick":
                self.controller = SteeringWheelController()
            elif config["controller"] == "keyboard":
                self.controller = KeyboardController(False)
            elif config["controller"] == "xboxController":
                self.controller = XboxController()
            else:
                raise ValueError("Unknown Policy: {}".format(config["controller"]))
        self.takeover = False

    def act(self, agent_id):
        agent_action = super(TakeoverPolicy, self).act(agent_id)
        if self.engine.global_config["manual_control"] and self.engine.agent_manager.get_agent(
                agent_id) is self.engine.current_track_vehicle and not self.engine.main_camera.is_bird_view_camera():
            expert_action = self.controller.process_input(self.engine.current_track_vehicle)
            if isinstance(self.controller, SteeringWheelController) and (self.controller.left_shift_paddle
                                                                         or self.controller.right_shift_paddle):
                # if expert_action[0]*agent_action[0]< 0 or expert_action[1]*agent_action[1] < 0:
                self.takeover = True
                return expert_action
            elif isinstance(self.controller, KeyboardController) and abs(sum(expert_action)) > 1e-2:
                self.takeover = True
                return expert_action
            elif isinstance(self.controller, XboxController) and (self.controller.button_x or self.controller.button_y):
                self.takeover = True
                return expert_action
        self.takeover = False
        return agent_action



import os
import pickle
import shutil

from metadrive.envs.real_data_envs.nuplan_env import NuPlanEnv
from metadrive.envs.scenario_env import ScenarioEnv
from metadrive.policy.replay_policy import NuPlanReplayEgoCarPolicy, ReplayEgoCarPolicy
from metadrive.scenario.utils import assert_scenario_equal


def _test_export_nuplan_scenario_hard(start_seed=0, num_scenarios=5, render_export_env=False, render_load_env=False):
    # ===== Save data =====
    env = NuPlanEnv(
        {
            "use_render": render_export_env,
            "agent_policy": NuPlanReplayEgoCarPolicy,
            "start_scenario_index": start_seed,
            "load_city_map": True,
            "num_scenarios": num_scenarios,
            "vehicle_config": dict(
                lidar=dict(num_lasers=120, distance=50, num_others=0),
                lane_line_detector=dict(num_lasers=12, distance=50),
                side_detector=dict(num_lasers=160, distance=50),
                # show_lidar=True
                show_navi_mark=False,
                show_dest_mark=False,
                no_wheel_friction=True,
            ),
        }
    )
    policy = lambda x: [0, 1]
    dir1 = None
    try:
        scenarios, done_info = env.export_scenarios(
            policy, scenario_index=[i for i in range(start_seed, start_seed + num_scenarios)]
        )
        dir1 = os.path.join(os.path.dirname(__file__), "test_export_nuplan_scenario_hard")
        os.makedirs(dir1, exist_ok=True)
        for i, data in scenarios.items():
            with open(os.path.join(dir1, "{}.pkl".format(i)), "wb+") as file:
                pickle.dump(data, file)
    finally:
        env.close()
        # pass

    # ===== Save data of the restoring environment =====
    env = ScenarioEnv(
        dict(
            agent_policy=ReplayEgoCarPolicy,
            data_directory=dir1,
            use_render=render_load_env,
            num_scenarios=num_scenarios,
            start_scenario_index=start_seed,
            force_reuse_object_name=True,
            # debug=True,
            # debug_static_world=True,
            vehicle_config=dict(no_wheel_friction=True)
        )
    )
    try:
        scenarios_restored, done_info = env.export_scenarios(
            policy, scenario_index=[i for i in range(num_scenarios)], render_topdown=False, return_done_info=True
        )
        for seed, info in done_info.items():
            if not info["arrive_dest"]:
                raise ValueError("Seed: {} Can not arrive dest!".format(seed))
    finally:
        env.close()

    if dir1 is not None:
        shutil.rmtree(dir1)

    assert_scenario_equal(scenarios, scenarios_restored, only_compare_sdc=False)


if __name__ == "__main__":
    _test_export_nuplan_scenario_hard(num_scenarios=5, render_export_env=False, render_load_env=True)



from metadrive.envs.metadrive_env import MetaDriveEnv


class MixedTrafficEnv(MetaDriveEnv):
    @classmethod
    def default_config(cls) -> "Config":
        config = super(MixedTrafficEnv, cls).default_config()
        config["rl_agent_ratio"] = 0.0
        return config

    def setup_engine(self):
        super(MetaDriveEnv, self).setup_engine()
        self.engine.accept("b", self.switch_to_top_down_view)
        self.engine.accept("q", self.switch_to_third_person_view)
        from metadrive.manager.traffic_manager import MixedPGTrafficManager
        from metadrive.manager.pg_map_manager import PGMapManager
        self.engine.register_manager("map_manager", PGMapManager())
        self.engine.register_manager("traffic_manager", MixedPGTrafficManager())


if __name__ == '__main__':
    env = MixedTrafficEnv(
        {
            "rl_agent_ratio": 0.5,
            "manual_control": True,
            # "use_render": True,
            "disable_model_compression": True,
            # "map": "SS",
            "num_scenarios": 100,
        }
    )
    try:
        obs, _ = env.reset()
        obs_space = env.observation_space
        assert obs_space.contains(obs)
        for _ in range(100000):
            assert env.observation_space.contains(obs)
            o, r, tm, tc, i = env.step(env.action_space.sample())
            assert obs_space.contains(o)
            if tm or tc:
                o, _ = env.reset()
                assert obs_space.contains(o)
    finally:
        env.close()



import logging
from collections import namedtuple

import numpy as np

from metadrive.component.map.base_map import BaseMap
from metadrive.component.vehicle.vehicle_type import *
from metadrive.manager.base_manager import BaseManager

BlockVehicles = namedtuple("block_vehicles", "trigger_road vehicles")


class RealDataManager(BaseManager):
    VEHICLE_GAP = 10  # m

    def __init__(self):
        """
        Replay Argoverse data.
        """
        super(RealDataManager, self).__init__()
        self._traffic_vehicles = []

    def reset(self):
        """
        Generate traffic on map, according to the mode and density
        :return: List of Traffic vehicles
        """
        map = self.current_map
        self._create_argoverse_vehicles_once(map)

    def before_step(self):
        """
        All traffic vehicles make driving decision here
        :return: None
        """
        # trigger vehicles
        engine = self.engine

        for v in self._traffic_vehicles:
            p = self.engine.get_policy(v.name)
            v.before_step(p.act())
        return dict()

    def before_reset(self) -> None:
        """
        Clear the scene and then reset the scene to empty
        :return: None
        """
        super(RealDataManager, self).before_reset()
        self.density = self.engine.global_config["traffic_density"]
        self._traffic_vehicles = []

    def get_vehicle_num(self):
        """
        Get the vehicles on road
        :return:
        """
        return len(self._traffic_vehicles)

    def _create_argoverse_vehicles_once(self, map: BaseMap) -> None:
        """
        Trigger mode, vehicles will be triggered only once, and disappear when arriving destination
        :param map: Map map.road_network[index]
        :param traffic_density: it can be adjusted each episode
        :return: None
        """
        real_data_config = self.engine.global_config["real_data_config"]
        locate_info = real_data_config["locate_info"]
        pos_dict = {i: j["init_pos"] for i, j in zip(locate_info.keys(), locate_info.values())}

        block = map.blocks[0]
        lanes = block.argo_lanes.values()
        roads = self.get_roads(block.block_network, direction='positive', lane_num=1)
        potential_vehicle_configs = []
        for l in lanes:
            start = np.max(l.centerline, axis=0)
            end = np.min(l.centerline, axis=0)
            for idx, pos in zip(pos_dict.keys(), pos_dict.values()):
                v_type = self.random_vehicle_type(prob=[0.4, 0.3, 0.3, 0, 0])
                if start[0] > pos[0] > end[0] and start[1] > pos[1] > end[1]:
                    long, lat = l.local_coordinates(pos)
                    config = {
                        "id": idx,
                        "type": v_type,
                        "v_config": {
                            "spawn_lane_index": l.index,
                            "spawn_longitude": long,
                            "enable_reverse": False,
                        }
                    }
                    potential_vehicle_configs.append(config)
                    pos_dict.pop(idx, None)
                    break
        from metadrive.policy.replay_policy import ReplayPolicy
        for road in roads:
            for config in potential_vehicle_configs:
                v_config = config["v_config"]
                v_start = v_config["spawn_lane_index"][0]
                v_end = v_config["spawn_lane_index"][1]
                v_config.update(self.engine.global_config["traffic_vehicle_config"])
                if road.start_node == v_start and road.end_node == v_end:
                    generated_v = self.spawn_object(config["type"], vehicle_config=v_config)
                    generated_v.set_static(True)
                    self.add_policy(generated_v.id, ReplayPolicy, generated_v, locate_info[config["id"]])
                    self._traffic_vehicles.append(generated_v)
                    potential_vehicle_configs.remove(config)
                    break

    def random_vehicle_type(self, prob=[0.2, 0.3, 0.3, 0.2, 0]):
        from metadrive.component.vehicle.vehicle_type import random_vehicle_type
        vehicle_type = random_vehicle_type(self.np_random, prob)
        return vehicle_type

    def destroy(self) -> None:
        """
        Destory func, release resource
        :return: None
        """
        self.clear_objects([v.id for v in self._traffic_vehicles])
        self._traffic_vehicles = []
        # current map

        # traffic vehicle list
        self._traffic_vehicles = None

    def __del__(self):
        logging.debug("{} is destroyed".format(self.__class__.__name__))

    def __repr__(self):
        return self.vehicles.__repr__()

    @property
    def vehicles(self):
        return list(self.engine.get_objects(filter=lambda o: isinstance(o, BaseVehicle)).values())

    @property
    def traffic_vehicles(self):
        return list(self._traffic_vehicles)

    @property
    def current_map(self):
        return self.engine.map_manager.current_map

    @staticmethod
    def get_roads(roadnetwork, *, direction="all", lane_num=None) -> List:
        """
        Return all roads in road_network
        :param direction: "positive"/"negative"
        :param lane_num: only roads with lane_num lanes will be returned
        :return: List[Road]
        """
        assert direction in ["positive", "negative", "all"], "incorrect road direction"
        ret = []
        for _from, _to_dict in self.graph.items():
            if direction == "all" or (direction == "positive" and _from[0] != "-") or (direction == "negative"
                                                                                       and _from[0] == "-"):
                for _to, lanes in _to_dict.items():
                    if lane_num is None or len(lanes) == lane_num:
                        ret.append(Road(_from, _to))
        return ret



from metadrive.component.sensors.base_camera import BaseCamera
from metadrive.constants import CamMask
from metadrive.engine.engine_utils import engine_initialized, get_global_config
from direct.filter.CommonFilters import CommonFilters


class RGBCamera(BaseCamera):
    # shape(dim_1, dim_2)
    BUFFER_W = 84  # dim 1
    BUFFER_H = 84  # dim 2
    CAM_MASK = CamMask.RgbCam
    PBR_ADAPT = False

    def __init__(self, engine, width, height, *, cuda=False):
        self.BUFFER_W, self.BUFFER_H = width, height
        super(RGBCamera, self).__init__(engine, True, cuda)
        cam = self.get_cam()
        lens = self.get_lens()
        # cam.lookAt(0, 2.4, 1.3)
        cam.lookAt(0, 10.4, 1.6)

        lens.setFov(60)
        # lens.setAspectRatio(2.0)



from metadrive.envs.metadrive_env import MetaDriveEnv

if __name__ == "__main__":
    env = MetaDriveEnv(
        {
            "num_scenarios": 1,
            "traffic_density": 0.1,
            "start_seed": 4,
            "vehicle_config": {
                "stack_size": 5,
                "rgb_clip": False,
                "rgb_camera": (800, 500),
                "rgb_to_grayscale": True
            },
            "manual_control": True,
            "use_render": False,
            "image_observation": True,  # it is a switch telling metadrive to use rgb as observation
            "rgb_clip": False,  # clip rgb to range(0,1) instead of (0, 255)
            # "pstats": True,
        }
    )
    env.reset()
    # # print m to capture rgb observation
    env.engine.accept(
        "m", env.vehicle.get_camera(env.vehicle.config["image_source"]).save_image, extraArgs=[env.vehicle]
    )
    import cv2

    for i in range(1, 100000):
        o, r, tm, tc, info = env.step([0, 1])
        assert env.observation_space.contains(o)
        # save
        rgb_cam = env.vehicle.get_camera(env.vehicle.config["image_source"])
        rgb_cam.save_image(env.vehicle, name="{}.png".format(i))
        cv2.imshow('img', o["image"][..., -1] / 255)
        cv2.waitKey(0)

        # if env.config["use_render"]:
        # for i in range(ImageObservation.STACK_SIZE):
        #      ObservationType.show_gray_scale_array(o["image"][:, :, i])
        # image = env.render(mode="any str except human", text={"can you see me": i})
        # ObservationType.show_gray_scale_array(image)
        if tm or tc:
            # print("Reset")
            env.reset()
    env.close()



import copy
import logging
from collections import OrderedDict
from typing import Union, List

from metadrive.component.block.base_block import BaseBlock
from metadrive.component.road_network import Road
from metadrive.component.road_network.node_road_network import NodeRoadNetwork


class PGBlockSocket:
    """
    A pair of roads in reverse direction
    Positive_road is right road, and Negative road is left road on which cars drive in reverse direction
    BlockSocket is a part of block used to connect other blocks
    """
    def __init__(self, positive_road: Road, negative_road: Road = None):
        self.positive_road = positive_road
        self.negative_road = negative_road if negative_road else None
        self.index = None

    def set_index(self, block_name: str, index: int):
        self.index = self.get_real_index(block_name, index)

    @classmethod
    def get_real_index(cls, block_name: str, index: int):
        return "{}-socket{}".format(block_name, index)

    def is_socket_node(self, road_node):
        if road_node == self.positive_road.start_node or road_node == self.positive_road.end_node or \
                road_node == self.negative_road.start_node or road_node == self.negative_road.end_node:
            return True
        else:
            return False

    def get_socket_in_reverse(self):
        """
        Return a new socket whose positive road=self.negative_road, negative_road=self.positive_road
        """
        new_socket = copy.deepcopy(self)
        new_socket.positive_road, new_socket.negative_road = self.negative_road, self.positive_road
        return new_socket

    def is_same_socket(self, other):
        return True if self.positive_road == other.positive_road and self.negative_road == other.negative_road else False

    def get_positive_lanes(self, global_network):
        return self.positive_road.get_lanes(global_network)

    def get_negative_lanes(self, global_network):
        return self.negative_road.get_lanes(global_network)


class PGBlock(BaseBlock):
    """
    Abstract class of Block,
    BlockSocket: a part of previous block connecting this block

    <----------------------------------------------
    road_2_end <---------------------- road_2_start
    <~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~>
    road_1_start ----------------------> road_1_end
    ---------------------------------------------->
    BlockSocket = tuple(road_1, road_2)

    When single-direction block created, road_2 in block socket is useless.
    But it's helpful when a town is created.
    """
    def __init__(
        self,
        block_index: int,
        pre_block_socket: PGBlockSocket,
        global_network: NodeRoadNetwork,
        random_seed,
        ignore_intersection_checking=False
    ):

        self.name = str(block_index) + self.ID
        super(PGBlock, self).__init__(
            block_index, global_network, random_seed, ignore_intersection_checking=ignore_intersection_checking
        )
        # block information
        assert self.SOCKET_NUM is not None, "The number of Socket should be specified when define a new block"
        if block_index == 0:
            from metadrive.component.pgblock.first_block import FirstPGBlock
            assert isinstance(self, FirstPGBlock), "only first block can use block index 0"
        elif block_index < 0:
            logging.debug("It is recommended that block index should > 1")
        self.number_of_sample_trial = 0

        # own sockets, one block derives from a socket, but will have more sockets to connect other blocks
        self._sockets = OrderedDict()

        # used to connect previous blocks, save its info here
        self.pre_block_socket = pre_block_socket
        self.pre_block_socket_index = pre_block_socket.index

        # used to create this block, but for first block it is nonsense
        if block_index != 0:
            self.positive_lanes = self.pre_block_socket.get_positive_lanes(self._global_network)
            self.negative_lanes = self.pre_block_socket.get_negative_lanes(self._global_network)
            self.positive_lane_num = len(self.positive_lanes)
            self.negative_lane_num = len(self.negative_lanes)
            self.positive_basic_lane = self.positive_lanes[-1]  # most right or outside lane is the basic lane
            self.negative_basic_lane = self.negative_lanes[-1]  # most right or outside lane is the basic lane
            self.lane_width = self.positive_basic_lane.width_at(0)

    def _sample_topology(self) -> bool:
        """
        Sample a new topology, clear the previous settings at first
        """
        self.number_of_sample_trial += 1
        no_cross = self._try_plug_into_previous_block()
        return no_cross

    def get_socket(self, index: Union[str, int]) -> PGBlockSocket:
        if isinstance(index, int):
            if index < 0 or index >= len(self._sockets):
                raise ValueError("Socket of {}: index out of range".format(self.class_name))
            socket_index = list(self._sockets)[index]
        else:
            assert index.startswith(self.name)
            socket_index = index
        assert socket_index in self._sockets, (socket_index, self._sockets.keys())
        return self._sockets[socket_index]

    def add_respawn_roads(self, respawn_roads: Union[List[Road], Road]):
        """
        Use this to add spawn roads instead of modifying the list directly
        """
        if isinstance(respawn_roads, List):
            for road in respawn_roads:
                self._add_one_respawn_road(road)
        elif isinstance(respawn_roads, Road):
            self._add_one_respawn_road(respawn_roads)
        else:
            raise ValueError("Only accept List[Road] or Road in this func")

    def add_sockets(self, sockets: Union[List[PGBlockSocket], PGBlockSocket]):
        """
        Use this to add sockets instead of modifying the list directly
        """
        if isinstance(sockets, PGBlockSocket):
            self._add_one_socket(sockets)
        elif isinstance(sockets, List):
            for socket in sockets:
                self._add_one_socket(socket)

    def _add_one_socket(self, socket: PGBlockSocket):
        assert isinstance(socket, PGBlockSocket), "Socket list only accept BlockSocket Type"
        if socket.index is not None and not socket.index.startswith(self.name):
            logging.warning(
                "The adding socket has index {}, which is not started with this block name {}. This is dangerous! "
                "Current block has sockets: {}.".format(socket.index, self.name, self.get_socket_indices())
            )
        if socket.index is None:
            # if this socket is self block socket
            socket.set_index(self.name, len(self._sockets))
        self._sockets[socket.index] = socket

    def _clear_topology(self):
        super(PGBlock, self)._clear_topology()
        self._sockets.clear()

    def _try_plug_into_previous_block(self) -> bool:
        """
        Try to plug this Block to previous block's socket, return True for success, False for road cross
        """
        raise NotImplementedError

    @staticmethod
    def create_socket_from_positive_road(road: Road) -> PGBlockSocket:
        """
        We usually create road from positive road, thus this func can get socket easily.
        Note: it is not recommended to generate socket from negative road
        """
        assert road.start_node[0] != Road.NEGATIVE_DIR and road.end_node[0] != Road.NEGATIVE_DIR, \
            "Socket can only be created from positive road"
        positive_road = Road(road.start_node, road.end_node)
        return PGBlockSocket(positive_road, -positive_road)

    def get_socket_indices(self):
        ret = list(self._sockets.keys())
        for r in ret:
            assert isinstance(r, str)
        return ret

    def get_socket_list(self):
        return list(self._sockets.values())

    def set_part_idx(self, x):
        """
        It is necessary to divide block to some parts in complex block and give them unique id according to part idx
        """
        self.PART_IDX = x
        self.ROAD_IDX = 0  # clear the road idx when create new part

    def add_road_node(self):
        """
        Call me to get a new node name of this block.
        It is more accurate and recommended to use road_node() to get a node name
        """
        self.ROAD_IDX += 1
        return self.road_node(self.PART_IDX, self.ROAD_IDX - 1)

    def road_node(self, part_idx: int, road_idx: int) -> str:
        """
        return standard road node name
        """
        return self.node(self.block_index, part_idx, road_idx)

    @classmethod
    def node(cls, block_idx: int, part_idx: int, road_idx: int) -> str:
        return str(block_idx) + cls.ID + str(part_idx) + cls.DASH + str(road_idx) + cls.DASH

    def get_intermediate_spawn_lanes(self):
        trigger_lanes = self.block_network.get_positive_lanes()
        respawn_lanes = self.get_respawn_lanes()
        for lanes in respawn_lanes:
            if lanes not in trigger_lanes:
                trigger_lanes.append(lanes)
        return trigger_lanes

    @property
    def block_network_type(self):
        return NodeRoadNetwork

    def create_in_world(self):
        graph = self.block_network.graph
        for _from, to_dict in graph.items():
            for _to, lanes in to_dict.items():
                for _id, lane in enumerate(lanes):
                    lane.construct_lane_in_block(self, (_from, _to, _id))
                    pos_road = not Road(_from, _to).is_negative_road()
                    lane.construct_lane_line_in_block(self, [True, True] if _id == 0 and pos_road else [False, True])



import logging
from typing import Optional

from metadrive.engine.base_engine import BaseEngine
from metadrive.engine.core.engine_core import EngineCore


def initialize_engine(env_global_config):
    cls = BaseEngine
    if cls.singleton is None:
        # assert cls.global_config is not None, "Set global config before initialization BaseEngine"
        cls.singleton = cls(env_global_config)
    else:
        raise PermissionError("There should be only one BaseEngine instance in one process")
    return cls.singleton


def get_engine() -> BaseEngine:
    return BaseEngine.singleton


def get_object(object_name):
    return get_engine().get_objects([object_name])


def engine_initialized():
    return False if BaseEngine.singleton is None else True


def close_engine():
    if BaseEngine.singleton is not None:
        BaseEngine.singleton.close()
        BaseEngine.singleton = None


def get_global_config():
    return EngineCore.global_config


def initialize_global_config(global_config):
    """
    You can, of course, preset the engine config before launching the engine.
    """
    assert not engine_initialized(), "Can not call this API after engine initialization!"
    EngineCore.global_config = global_config


def set_global_random_seed(random_seed: Optional[int]):
    """
    Update the random seed and random engine
    All subclasses of Randomizable will hold the same random engine, after calling this function
    :param random_seed: int, random seed
    """
    engine = get_engine()
    if engine is not None:
        engine.seed(random_seed)
    else:
        logging.warning("BaseEngine is not launched, fail to sync seed to engine!")



"""

Converts pipeline resources to TXO files to speed up loading.
This is called during the setup.

"""

from __future__ import print_function

import os

from panda3d.core import Filename, Texture, load_prc_file_data
load_prc_file_data("", "window-type none")
load_prc_file_data("", "notify-level-pnmimage error")
load_prc_file_data("", "textures-power-2 none")

files_to_convert = [
    "data/gui/loading_screen_bg.png", "rpplugins/bloom/resources/lens_dirt.png", "data/builtin_models/skybox/skybox.jpg"
]

this_dir = os.path.realpath(os.path.dirname(__file__))
os.chdir(this_dir)

pipeline_dir = "../"

import direct.directbase.DirectStart

for filename in files_to_convert:
    src_path = os.path.abspath(os.path.join(pipeline_dir, filename))
    fullpath = Filename.from_os_specific(src_path).get_fullpath()
    dest_path = fullpath.replace(".png", ".txo.pz")
    dest_path = dest_path.replace(".jpg", ".txo.pz")

    print(src_path, "->", dest_path)
    loader.load_texture(fullpath).write(dest_path)



"""
This script demonstrates how to setup the Multi-agent RL environments.


Usage: python -m metadrive.examples.drive_in_multi_agent_env --env pgma

Options for --env argument:
    (1) roundabout
    (2) intersection
    (3) tollgate
    (4) bottleneck
    (5) parkinglot
    (6) pgma

"""
import argparse
from metadrive.component.sensors.rgb_camera import RGBCamera
from metadrive import (
    MultiAgentMetaDrive, MultiAgentTollgateEnv, MultiAgentBottleneckEnv, MultiAgentIntersectionEnv,
    MultiAgentRoundaboutEnv, MultiAgentParkingLotEnv
)
from metadrive.constants import HELP_MESSAGE
from metadrive.policy.idm_policy import ManualControllableIDMPolicy

if __name__ == "__main__":
    envs = dict(
        roundabout=MultiAgentRoundaboutEnv,
        intersection=MultiAgentIntersectionEnv,
        tollgate=MultiAgentTollgateEnv,
        bottleneck=MultiAgentBottleneckEnv,
        parkinglot=MultiAgentParkingLotEnv,
        pgma=MultiAgentMetaDrive
    )

    parser = argparse.ArgumentParser()
    parser.add_argument("--env", type=str, default="roundabout", choices=list(envs.keys()))
    parser.add_argument("--top_down", action="store_true")
    args = parser.parse_args()
    env_cls_name = args.env
    extra_args = dict(film_size=(800, 800)) if args.top_down else {}
    assert env_cls_name in envs.keys(), "No environment named {}, argument accepted: \n" \
                                        "(1) roundabout\n" \
                                        "(2) intersection\n" \
                                        "(3) tollgate\n" \
                                        "(4) bottleneck\n" \
                                        "(5) parkinglot\n" \
                                        "(6) pgma" \
        .format(env_cls_name)
    env = envs[env_cls_name](
        {
            "use_render": True if not args.top_down else False,
            "crash_done": False,
            "sensors": dict(rgb_camera=(RGBCamera, 512, 256)),
            "interface_panel": ["rgb_camera", "dashboard"],
            "agent_policy": ManualControllableIDMPolicy
        }
    )
    try:
        env.reset()
        # if env.current_track_vehicle:
        #     env.current_track_vehicle.expert_takeover = True
        print(HELP_MESSAGE)
        env.switch_to_third_person_view()  # Default is in Top-down view, we switch to Third-person view.
        for i in range(1, 10000000000):
            o, r, tm, tc, info = env.step({agent_id: [0, 0] for agent_id in env.vehicles.keys()})
            env.render(
                **extra_args,
                mode="top_down" if args.top_down else None,
                text={
                    "Quit": "ESC",
                    "Number of existing vehicles": len(env.vehicles),
                    "Tracked agent (Press Q)": env.engine.agent_manager.object_to_agent(env.current_track_vehicle.id),
                    "Keyboard Control": "W,A,S,D",
                    # "Auto-Drive (Switch mode: T)": "on" if env.current_track_vehicle.expert_takeover else "off",
                } if not args.top_down else {}
            )
            if tm["__all__"]:
                env.reset()
                # if env.current_track_vehicle:
                #     env.current_track_vehicle.expert_takeover = True
    except Exception as e:
        raise e
    finally:
        env.close()



from metadrive.envs.marl_envs.marl_inout_roundabout import MultiAgentRoundaboutEnv
from metadrive.utils import setup_logger


class TestEnv(MultiAgentRoundaboutEnv):
    def __init__(self):
        super(TestEnv, self).__init__(
            config={
                "use_render": True,
                "map": "SSS",
                "num_agents": 4,
                "force_destroy": True,
                "manual_control": True,
                "target_vehicle_configs": {"agent{}".format(i): {
                    "spawn_longitude": i * 5
                }
                                           for i in range(4)}
            }
        )


if __name__ == "__main__":
    setup_logger(True)
    env = TestEnv()

    o, _ = env.reset()
    # print("vehicle num", len(env.engine.traffic_manager.vehicles))
    for i in range(1, 100000):
        o, r, tm, tc, info = env.step({key: [0, 0] for key in env.action_space.sample()})
        # o, r, tm, tc, info = env.step([0,1])
        env.render(text={"display_regions": len(env.engine.win.getDisplayRegions())})
        if True in tm.values() or True in tc.values():
            # print("Reset")
            env.reset()
    env.close()



from metadrive.envs.marl_envs import MultiAgentRoundaboutEnv


def test_engine_buffer():
    env = MultiAgentRoundaboutEnv({"num_buffering_objects": 0, "debug": True})
    try:
        for i in range(1, 30):
            env.reset()
    finally:
        env.close()
    env = MultiAgentRoundaboutEnv({"num_buffering_objects": 200, "debug": True})
    try:
        for i in range(1, 30):
            env.reset()
    finally:
        env.close()


if __name__ == '__main__':
    test_engine_buffer()
